# SPDX-License-Identifier: LGPL-2.1-or-later

# /***************************************************************************
# *   Copyright (c) 2016 Victor Titov (DeepSOIC) <vv.titov@gmail.com>       *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This library is free software; you can redistribute it and/or         *
# *   modify it under the terms of the GNU Library General Public           *
# *   License as published by the Free Software Foundation; either          *
# *   version 2 of the License, or (at your option) any later version.      *
# *                                                                         *
# *   This library  is distributed in the hope that it will be useful,      *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this library; see the file COPYING.LIB. If not,    *
# *   write to the Free Software Foundation, Inc., 59 Temple Place,         *
# *   Suite 330, Boston, MA  02111-1307, USA                                *
# *                                                                         *
# ***************************************************************************/

__title__ = "BOPTools.SplitFeatures module"
__author__ = "DeepSOIC"
__url__ = "https://www.freecad.org"
__doc__ = "Shape splitting document objects (features)."

from . import SplitAPI
import FreeCAD
import Part

if FreeCAD.GuiUp:
    import FreeCADGui
    from PySide import QtCore, QtGui

    # -------------------------- translation-related code -------------------------
    # See forum thread "A new Part tool is being born... JoinFeatures!"
    # https://forum.freecad.org/viewtopic.php?f=22&t=11112&start=30#p90239
    try:
        _fromUtf8 = QtCore.QString.fromUtf8
    except Exception:

        def _fromUtf8(s):
            return s

    translate = FreeCAD.Qt.translate
# --------------------------/translation-related code --------------------------


def getIconPath(icon_dot_svg):
    return icon_dot_svg


# -------------------------- /common stuff ------------------------------------

# -------------------------- BooleanFragments ---------------------------------


def makeBooleanFragments(name):
    """makeBooleanFragments(name): makes an BooleanFragments object."""
    obj = FreeCAD.ActiveDocument.addObject("Part::FeaturePython", name)
    FeatureBooleanFragments(obj)
    if FreeCAD.GuiUp:
        ViewProviderBooleanFragments(obj.ViewObject)
    return obj


class FeatureBooleanFragments:
    """The BooleanFragments feature object."""

    def __init__(self, obj):
        obj.addProperty(
            "App::PropertyLinkList",
            "Objects",
            "BooleanFragments",
            "Object to compute intersections between.",
            locked=True,
        )
        obj.addProperty(
            "App::PropertyEnumeration",
            "Mode",
            "BooleanFragments",
            "- Standard: wires, shells, compsolids remain in one piece.\n"
            "- Split: wires, shells, compsolids are split.\n"
            "- CompSolid: make compsolid from solid fragments.",
            locked=True,
        )
        obj.Mode = ["Standard", "Split", "CompSolid"]
        obj.addProperty(
            "App::PropertyLength",
            "Tolerance",
            "BooleanFragments",
            "Tolerance when intersecting (fuzzy value). "
            "In addition to tolerances of the shapes.",
            locked=True,
        )

        obj.Proxy = self
        self.Type = "FeatureBooleanFragments"

    def execute(self, selfobj):
        shapes = [obj.Shape for obj in selfobj.Objects]
        if len(shapes) == 1 and shapes[0].ShapeType == "Compound":
            shapes = shapes[0].childShapes()
        if len(shapes) < 2:
            raise ValueError(
                "At least two shapes are needed for computing boolean fragments. Got only {num}.".format(
                    num=len(shapes)
                )
            )
        selfobj.Shape = SplitAPI.booleanFragments(shapes, selfobj.Mode, selfobj.Tolerance)


class ViewProviderBooleanFragments:
    """A View Provider for the Part BooleanFragments feature."""

    def __init__(self, vobj):
        vobj.Proxy = self

    def getIcon(self):
        return ":/icons/booleans/Part_BooleanFragments.svg"

    def attach(self, vobj):
        self.ViewObject = vobj
        self.Object = vobj.Object

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def claimChildren(self):
        return self.Object.Objects

    def onDelete(self, feature, subelements):
        try:
            for obj in self.claimChildren():
                obj.ViewObject.show()
        except Exception as err:
            FreeCAD.Console.PrintError("Error in onDelete: " + str(err))
        return True

    def canDragObjects(self):
        return True

    def canDropObjects(self):
        return True

    def canDragObject(self, dragged_object):
        return True

    def canDropObject(self, incoming_object):
        return hasattr(incoming_object, "Shape")

    def dragObject(self, selfvp, dragged_object):
        objs = self.Object.Objects
        objs.remove(dragged_object)
        self.Object.Objects = objs

    def dropObject(self, selfvp, incoming_object):
        self.Object.Objects = self.Object.Objects + [incoming_object]


def cmdCreateBooleanFragmentsFeature(name, mode):
    """cmdCreateBooleanFragmentsFeature(name, mode): implementation of GUI command to create
    BooleanFragments feature (GFA). Mode can be "Standard", "Split", or "CompSolid"."""
    sel = FreeCADGui.Selection.getSelectionEx()
    FreeCAD.ActiveDocument.openTransaction("Create Boolean Fragments")
    FreeCADGui.addModule("BOPTools.SplitFeatures")
    FreeCADGui.doCommand(
        "j = BOPTools.SplitFeatures.makeBooleanFragments(name='{name}')".format(name=name)
    )
    FreeCADGui.doCommand(
        "j.Objects = {sel}".format(
            sel="[" + ", ".join(["App.ActiveDocument." + so.Object.Name for so in sel]) + "]"
        )
    )
    FreeCADGui.doCommand("j.Mode = {mode}".format(mode=repr(mode)))

    try:
        FreeCADGui.doCommand("j.Proxy.execute(j)")
        FreeCADGui.doCommand("j.purgeTouched()")
    except Exception as err:
        mb = QtGui.QMessageBox()
        mb.setIcon(mb.Icon.Warning)
        error_text1 = translate("Part_SplitFeatures", "Computing the result failed with an error:")
        error_text2 = translate(
            "Part_SplitFeatures",
            "Click 'Continue' to create the feature anyway, or 'Abort' to cancel.",
        )
        mb.setText(error_text1 + "\n\n" + str(err) + "\n\n" + error_text2)
        mb.setWindowTitle(translate("Part_SplitFeatures", "Bad Selection", None))
        btnAbort = mb.addButton(QtGui.QMessageBox.StandardButton.Abort)
        btnOK = mb.addButton(
            translate("Part_SplitFeatures", "Continue", None),
            QtGui.QMessageBox.ButtonRole.ActionRole,
        )
        mb.setDefaultButton(btnOK)

        mb.exec_()

        if mb.clickedButton() is btnAbort:
            FreeCAD.ActiveDocument.abortTransaction()
            return

    FreeCADGui.doCommand(
        "for obj in j.ViewObject.Proxy.claimChildren():\n" "    obj.ViewObject.hide()"
    )

    FreeCAD.ActiveDocument.commitTransaction()


class CommandBooleanFragments:
    """Command to create BooleanFragments feature."""

    def GetResources(self):
        return {
            "Pixmap": getIconPath("Part_BooleanFragments.svg"),
            "MenuText": QtCore.QT_TRANSLATE_NOOP("Part_BooleanFragments", "Boolean Fragments"),
            "Accel": "",
            "ToolTip": QtCore.QT_TRANSLATE_NOOP(
                "Part_BooleanFragments",
                "Creates a boolean union which is sliced at the intersections of the selected shapes",
            ),
        }

    def Activated(self):
        if len(FreeCADGui.Selection.getSelectionEx()) >= 1:
            cmdCreateBooleanFragmentsFeature(name="BooleanFragments", mode="Standard")
        else:
            mb = QtGui.QMessageBox()
            mb.setIcon(mb.Icon.Warning)
            mb.setText(
                translate(
                    "Part_SplitFeatures",
                    "Select at least two objects, or one or more compounds. "
                    "If only one compound is selected, the compounded shapes will be intersected between each other "
                    "(otherwise, compounds with self-intersections are invalid).",
                    None,
                )
            )
            mb.setWindowTitle(translate("Part_SplitFeatures", "Bad Selection", None))
            mb.exec_()

    def IsActive(self):
        if FreeCAD.ActiveDocument:
            return True
        else:
            return False


# -------------------------- /BooleanFragments --------------------------------

# -------------------------- Slice --------------------------------------------


def makeSlice(name):
    """makeSlice(name): makes an Slice object."""
    obj = FreeCAD.ActiveDocument.addObject("Part::FeaturePython", name)
    FeatureSlice(obj)
    if FreeCAD.GuiUp:
        ViewProviderSlice(obj.ViewObject)
    return obj


class FeatureSlice:
    """The Slice feature object."""

    def __init__(self, obj):
        obj.addProperty("App::PropertyLink", "Base", "Slice", "Object to be sliced.", locked=True)
        obj.addProperty(
            "App::PropertyLinkList", "Tools", "Slice", "Objects that slice.", locked=True
        )
        obj.addProperty(
            "App::PropertyEnumeration",
            "Mode",
            "Slice",
            "- Standard: wires, shells, compsolids remain in one piece.\n"
            "- Split: wires, shells, compsolids are split.\n"
            "- CompSolid: make compsolid from solid fragments.",
            locked=True,
        )
        obj.Mode = ["Standard", "Split", "CompSolid"]
        obj.addProperty(
            "App::PropertyLength",
            "Tolerance",
            "Slice",
            "Tolerance when intersecting (fuzzy value). "
            "In addition to tolerances of the shapes.",
            locked=True,
        )

        obj.Proxy = self
        self.Type = "FeatureSlice"

    def execute(self, selfobj):
        if len(selfobj.Tools) < 1:
            raise ValueError("No slicing objects supplied!")

        # helper function to get the shape from object or group
        def get_shape(obj):
            """get shape from a part object or compound from a group."""
            if hasattr(obj, "Shape"):
                return obj.Shape
            elif hasattr(obj, "Group"):  # it's a group/container from ie. slice apart
                shapes = []
                for child in obj.Group:
                    if hasattr(child, "Shape"):
                        shapes.append(child.Shape)
                if shapes:
                    return Part.makeCompound(shapes)
            return None

        # get base shape
        base_shape = get_shape(selfobj.Base)
        if base_shape is None or base_shape.isNull():
            raise ValueError("Base object has no valid shape!")

        # get tool shapes
        tool_shapes = []
        for tool in selfobj.Tools:
            shape = get_shape(tool)
            if shape is not None and not shape.isNull():
                tool_shapes.append(shape)

        if len(tool_shapes) < 1:
            raise ValueError("No valid tool shapes available")

        selfobj.Shape = SplitAPI.slice(
            base_shape,
            tool_shapes,
            selfobj.Mode,
            selfobj.Tolerance,
        )


class ViewProviderSlice:
    """A View Provider for the Part Slice feature."""

    def __init__(self, vobj):
        vobj.Proxy = self

    def getIcon(self):
        return ":/icons/booleans/Part_Slice.svg"

    def attach(self, vobj):
        self.ViewObject = vobj
        self.Object = vobj.Object

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def claimChildren(self):
        return [self.Object.Base] + self.Object.Tools

    def onDelete(self, feature, subelements):
        try:
            for obj in self.claimChildren():
                obj.ViewObject.show()
        except Exception as err:
            FreeCAD.Console.PrintError("Error in onDelete: " + str(err))
        return True


def cmdCreateSliceFeature(name, mode, transaction=True):
    """cmdCreateSliceFeature(name, mode): implementation of GUI command to create
    Slice feature. Mode can be "Standard", "Split", or "CompSolid"."""
    sel = FreeCADGui.Selection.getSelectionEx()
    if transaction:
        FreeCAD.ActiveDocument.openTransaction("Create Slice")
    FreeCADGui.addModule("BOPTools.SplitFeatures")
    FreeCADGui.doCommand("f = BOPTools.SplitFeatures.makeSlice(name='{name}')".format(name=name))
    FreeCADGui.doCommand(
        "f.Base = {sel}[0]\n"
        "f.Tools = {sel}[1:]".format(
            sel="[" + ", ".join(["App.ActiveDocument." + so.Object.Name for so in sel]) + "]"
        )
    )
    FreeCADGui.doCommand("f.Mode = {mode}".format(mode=repr(mode)))

    try:
        FreeCADGui.doCommand("f.Proxy.execute(f)")
        FreeCADGui.doCommand("f.purgeTouched()")
    except Exception as err:
        mb = QtGui.QMessageBox()
        mb.setIcon(mb.Icon.Warning)
        error_text1 = translate("Part_SplitFeatures", "Computing the result failed with an error:")
        error_text2 = translate(
            "Part_SplitFeatures",
            "Click 'Continue' to create the feature anyway, or 'Abort' to cancel.",
        )
        mb.setText(error_text1 + "\n\n" + str(err) + "\n\n" + error_text2)
        mb.setWindowTitle(translate("Part_SplitFeatures", "Bad Selection", None))
        btnAbort = mb.addButton(QtGui.QMessageBox.StandardButton.Abort)
        btnOK = mb.addButton(
            translate("Part_SplitFeatures", "Continue", None),
            QtGui.QMessageBox.ButtonRole.ActionRole,
        )
        mb.setDefaultButton(btnOK)

        mb.exec_()

        if mb.clickedButton() is btnAbort:
            if transaction:
                FreeCAD.ActiveDocument.abortTransaction()
            return False

    FreeCADGui.doCommand(
        "for obj in f.ViewObject.Proxy.claimChildren():\n" "    obj.ViewObject.hide()"
    )

    if transaction:
        FreeCAD.ActiveDocument.commitTransaction()
    return True


def cmdSliceApart():
    FreeCAD.ActiveDocument.openTransaction("Slice apart")
    made = cmdCreateSliceFeature(name="Slice", mode="Split", transaction=False)

    if made:
        FreeCADGui.addModule("CompoundTools.Explode")
        FreeCADGui.doCommand("CompoundTools.Explode.explodeCompound(f)")
        FreeCADGui.doCommand("f.ViewObject.hide()")
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCADGui.doCommand("App.ActiveDocument.recompute()")
    else:
        FreeCAD.ActiveDocument.abortTransaction()


class CommandSlice:
    """Command to create Slice feature."""

    def GetResources(self):
        return {
            "Pixmap": getIconPath("Part_Slice.svg"),
            "MenuText": QtCore.QT_TRANSLATE_NOOP("Part_Slice", "Slice to Compound"),
            "Accel": "",
            "ToolTip": QtCore.QT_TRANSLATE_NOOP(
                "Part_Slice",
                "Slices the selected object by using other objects as cutting tools and storing the results in one compound",
            ),
        }

    def Activated(self):
        if len(FreeCADGui.Selection.getSelectionEx()) > 1:
            cmdCreateSliceFeature(name="Slice", mode="Split")
        else:
            mb = QtGui.QMessageBox()
            mb.setIcon(mb.Icon.Warning)
            mb.setText(
                translate(
                    "Part_SplitFeatures",
                    "Select at least two objects. "
                    "The first one is the object to be sliced; "
                    "the rest are objects to slice with.",
                    None,
                )
            )
            mb.setWindowTitle(translate("Part_SplitFeatures", "Bad Selection", None))
            mb.exec_()

    def IsActive(self):
        if FreeCAD.ActiveDocument:
            return True
        else:
            return False


class CommandSliceApart:
    """Command to create exploded Slice feature."""

    def GetResources(self):
        return {
            "Pixmap": getIconPath("Part_SliceApart.svg"),
            "MenuText": QtCore.QT_TRANSLATE_NOOP("Part_SliceApart", "Slice Apart"),
            "Accel": "",
            "ToolTip": QtCore.QT_TRANSLATE_NOOP(
                "Part_SliceApart",
                "Slices the selected object by other objects, and splits it apart, creating a compound filter for each slide",
            ),
        }

    def Activated(self):
        if len(FreeCADGui.Selection.getSelectionEx()) > 1:
            cmdSliceApart()
        else:
            mb = QtGui.QMessageBox()
            mb.setIcon(mb.Icon.Warning)
            mb.setText(
                translate(
                    "Part_SplitFeatures",
                    "Select at least two objects. "
                    "The first one is the object to be sliced; "
                    "the rest are objects to slice with.",
                    None,
                )
            )
            mb.setWindowTitle(translate("Part_SplitFeatures", "Bad Selection", None))
            mb.exec_()

    def IsActive(self):
        if FreeCAD.ActiveDocument:
            return True
        else:
            return False


# -------------------------- /Slice -------------------------------------------

# -------------------------- XOR ----------------------------------------------


def makeXOR(name):
    """makeXOR(name): makes an XOR object."""
    obj = FreeCAD.ActiveDocument.addObject("Part::FeaturePython", name)
    FeatureXOR(obj)
    if FreeCAD.GuiUp:
        ViewProviderXOR(obj.ViewObject)
    return obj


class FeatureXOR:
    """The XOR feature object."""

    def __init__(self, obj):
        obj.addProperty(
            "App::PropertyLinkList",
            "Objects",
            "XOR",
            "Object to compute intersections between.",
            locked=True,
        )
        obj.addProperty(
            "App::PropertyLength",
            "Tolerance",
            "XOR",
            "Tolerance when intersecting (fuzzy value). "
            "In addition to tolerances of the shapes.",
            locked=True,
        )

        obj.Proxy = self
        self.Type = "FeatureXOR"

    def execute(self, selfobj):
        shapes = [obj.Shape for obj in selfobj.Objects]
        if len(shapes) == 1 and shapes[0].ShapeType == "Compound":
            shapes = shapes[0].childShapes()
        if len(shapes) < 2:
            raise ValueError(
                "At least two shapes are needed for computing XOR. Got only {num}.".format(
                    num=len(shapes)
                )
            )
        selfobj.Shape = SplitAPI.xor(shapes, selfobj.Tolerance)


class ViewProviderXOR:
    """A View Provider for the Part XOR feature."""

    def __init__(self, vobj):
        vobj.Proxy = self

    def getIcon(self):
        return ":/icons/booleans/Part_XOR.svg"

    def attach(self, vobj):
        self.ViewObject = vobj
        self.Object = vobj.Object

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def claimChildren(self):
        return self.Object.Objects

    def onDelete(self, feature, subelements):
        try:
            for obj in self.claimChildren():
                obj.ViewObject.show()
        except Exception as err:
            FreeCAD.Console.PrintError("Error in onDelete: " + str(err))
        return True

    def canDragObjects(self):
        return True

    def canDropObjects(self):
        return True

    def canDragObject(self, dragged_object):
        return True

    def canDropObject(self, incoming_object):
        return hasattr(incoming_object, "Shape")

    def dragObject(self, selfvp, dragged_object):
        objs = self.Object.Objects
        objs.remove(dragged_object)
        self.Object.Objects = objs

    def dropObject(self, selfvp, incoming_object):
        self.Object.Objects = self.Object.Objects + [incoming_object]


def cmdCreateXORFeature(name):
    """cmdCreateXORFeature(name): implementation of GUI command to create
    XOR feature (GFA). Mode can be "Standard", "Split", or "CompSolid"."""
    sel = FreeCADGui.Selection.getSelectionEx()
    FreeCAD.ActiveDocument.openTransaction("Create Boolean XOR")
    FreeCADGui.addModule("BOPTools.SplitFeatures")
    FreeCADGui.doCommand("j = BOPTools.SplitFeatures.makeXOR(name='{name}')".format(name=name))
    FreeCADGui.doCommand(
        "j.Objects = {sel}".format(
            sel="[" + ", ".join(["App.ActiveDocument." + so.Object.Name for so in sel]) + "]"
        )
    )

    try:
        FreeCADGui.doCommand("j.Proxy.execute(j)")
        FreeCADGui.doCommand("j.purgeTouched()")
    except Exception as err:
        mb = QtGui.QMessageBox()
        mb.setIcon(mb.Icon.Warning)
        error_text1 = translate("Part_SplitFeatures", "Computing the result failed with an error:")
        error_text2 = translate(
            "Part_SplitFeatures",
            "Click 'Continue' to create the feature anyway, or 'Abort' to cancel.",
        )
        mb.setText(error_text1 + "\n\n" + str(err) + "\n\n" + error_text2)
        mb.setWindowTitle(translate("Part_SplitFeatures", "Bad Selection", None))
        btnAbort = mb.addButton(QtGui.QMessageBox.StandardButton.Abort)
        btnOK = mb.addButton(
            translate("Part_SplitFeatures", "Continue", None),
            QtGui.QMessageBox.ButtonRole.ActionRole,
        )
        mb.setDefaultButton(btnOK)

        mb.exec_()

        if mb.clickedButton() is btnAbort:
            FreeCAD.ActiveDocument.abortTransaction()
            return

    FreeCADGui.doCommand(
        "for obj in j.ViewObject.Proxy.claimChildren():\n" "    obj.ViewObject.hide()"
    )

    FreeCAD.ActiveDocument.commitTransaction()


class CommandXOR:
    """Command to create XOR feature."""

    def GetResources(self):
        return {
            "Pixmap": getIconPath("Part_XOR.svg"),
            "MenuText": QtCore.QT_TRANSLATE_NOOP("Part_XOR", "Boolean XOR"),
            "Accel": "",
            "ToolTip": QtCore.QT_TRANSLATE_NOOP(
                "Part_XOR",
                "Performs an 'exclusive OR' boolean operation with two or more selected objects,\n"
                "or with the shapes inside a compound.\n"
                "Overlapping volumes of the shapes will be removed.",
            ),
        }

    def Activated(self):
        if len(FreeCADGui.Selection.getSelectionEx()) >= 1:
            cmdCreateXORFeature(name="XOR")
        else:
            mb = QtGui.QMessageBox()
            mb.setIcon(mb.Icon.Warning)
            mb.setText(
                translate(
                    "Part_SplitFeatures",
                    "Select at least two objects, or one or more compounds. "
                    "If only one compound is selected, the compounded shapes will be intersected between each other "
                    "(otherwise, compounds with self-intersections are invalid).",
                    None,
                )
            )
            mb.setWindowTitle(translate("Part_SplitFeatures", "Bad Selection", None))
            mb.exec_()

    def IsActive(self):
        if FreeCAD.ActiveDocument:
            return True
        else:
            return False


# -------------------------- /XOR ---------------------------------------------


def addCommands():
    FreeCADGui.addCommand("Part_BooleanFragments", CommandBooleanFragments())
    FreeCADGui.addCommand("Part_Slice", CommandSlice())
    FreeCADGui.addCommand("Part_SliceApart", CommandSliceApart())
    FreeCADGui.addCommand("Part_XOR", CommandXOR())



# ***************************************************************************
# *   Copyright (c) 2026 Chris Jones github.com/ipatch                      *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# ***************************************************************************

import FreeCAD
from CAMTests.PathTestUtils import PathTestWithAssets
from Path.Tool.toolbit import ToolBitEndmill


class TestToolBitRecomputeState(PathTestWithAssets):
    """Test toolbit recompute state handling (issue #26652)"""

    def setUp(self):
        """Set up test fixtures"""
        super().setUp()
        self.doc = FreeCAD.newDocument("TestToolbitState")

    def tearDown(self):
        """Clean up after tests"""
        FreeCAD.closeDocument(self.doc.Name)
        super().tearDown()

    def testToolbitExpressionRecomputeState(self):
        """Test that toolbit with unitless expression doesn't stay touched after recompute"""
        # create an endmill shape and toolbit
        shape = self.assets.get("toolbitshape://endmill")
        toolbit = ToolBitEndmill(shape, id="test_endmill")

        # attach to document
        obj = toolbit.attach_to_doc(self.doc, label="TestEndmill")

        # set initial diameter
        obj.Diameter = FreeCAD.Units.Quantity("5 mm")
        self.doc.recompute()

        # now set expression without unit - this triggers issue #26652
        obj.setExpression("Diameter", "5")
        self.doc.recompute()

        # after recompute, toolbit should not be in touched state
        self.assertFalse(
            "Touched" in obj.State,
            "Toolbit with unitless expression should not be touched after recompute",
        )



from Base.Metadata import export, constmethod
from Base.PyObjectBase import PyObjectBase
from typing import overload

@export(
    PythonName="Part.ShapeUpgrade.UnifySameDomain",
    Include="ShapeUpgrade_UnifySameDomain.hxx",
    Twin="ShapeUpgrade_UnifySameDomain",
    TwinPointer="ShapeUpgrade_UnifySameDomain",
    Constructor=True,
    Delete=True,
)
class UnifySameDomain(PyObjectBase):
    """
    This tool tries to unify faces and edges of the shape which lie on the same geometry.

    Author: Werner Mayer (wmayer@users.sourceforge.net)
    Licence: LGPL
    """

    def initialize(self, **kwargs) -> None:
        """
        Initializes with a shape and necessary flags
        """
        ...

    def allowInternalEdges(self) -> None:
        """
        Sets the flag defining whether it is allowed to create
        internal edges inside merged faces in the case of non-manifold
        topology. Without this flag merging through multi connected edge
        is forbidden. Default value is false.
        """
        ...

    def keepShape(self) -> None:
        """
        Sets the shape for avoid merging of the faces/edges.
        """
        ...

    def keepShapes(self) -> None:
        """
        Sets the map of shapes for avoid merging of the faces/edges.
        """
        ...

    def setSafeInputMode(self) -> None:
        """
        Sets the flag defining the behavior of the algorithm regarding
        modification of input shape.
        If this flag is equal to True then the input (original) shape can't be
        modified during modification process. Default value is true.
        """
        ...

    def setLinearTolerance(self) -> None:
        """
        Sets the linear tolerance
        """
        ...

    def setAngularTolerance(self) -> None:
        """
        Sets the angular tolerance
        """
        ...

    def build(self) -> None:
        """
        Performs unification and builds the resulting shape
        """
        ...

    @constmethod
    def shape(self) -> None:
        """
        Gives the resulting shape
        """
        ...



#!/usr/bin/env python3
import argparse
import sys
import os
import re
import logging
from typing import Tuple
from utils import (
    run_command,
    init_environment,
    append_file,
    emit_problem_matchers,
    write_file,
    add_common_arguments,
)

DEFAULT_LINE_LENGTH_LIMIT = 100


def parse_black_output(output: str) -> Tuple[int, int]:
    """
    Parse Black output to determine how many files would be reformatted and how many would fail.

    Returns:
        A tuple (black_reformats, black_fails).
    """
    black_reformats = 0
    black_fails = 0

    # Look for a pattern like: "X files would be reformatted"
    reformats_matches = re.findall(r"(\d+)(?=\s+fil.+ would be reformatted)", output)
    if reformats_matches:
        black_reformats = int(reformats_matches[0])

    # Look for a pattern like: "Y files would fail to reformat"
    fails_matches = re.findall(r"(\d+)(?=\s+fil.+ would fail to reformat)", output)
    if fails_matches:
        black_fails = int(fails_matches[0])

    return black_reformats, black_fails


def generate_markdown_report(black_reformats: int, black_fails: int, log_file: str) -> str:
    """Generate a Markdown report section based on Black results and log file."""
    report_lines = []
    if black_reformats > 0 or black_fails > 0:
        report_lines.append(
            f"<details><summary>:pencil2: Black would reformat {black_reformats} file(s)"
            f" and {black_fails} file(s) would fail to reformat</summary>"
        )
    else:
        report_lines.append(
            "<details><summary>:heavy_check_mark: Black would reformat no file</summary>"
        )
    report_lines.append("")
    report_lines.append("````")
    with open(log_file, "r", encoding="utf-8") as f:
        report_lines.append(f.read())
    report_lines.append("````")
    report_lines.append("</details>")
    report_lines.append("")
    return "\n".join(report_lines)


def main():
    parser = argparse.ArgumentParser(
        description="Run Black in check mode on Python files and append a Markdown report."
    )
    add_common_arguments(parser)
    args = parser.parse_args()
    init_environment(args)

    logging.info("Installing black (if needed)...")
    run_command(["pipx", "install", "-q", "black"], check=True)

    cmd = [
        "black",
        "--line-length",
        str(DEFAULT_LINE_LENGTH_LIMIT),
        "--check",
    ] + args.files.split()
    stdout, stderr, exit_code = run_command(cmd, check=False)
    output = stdout + "\n" + stderr

    log_file = os.path.join(args.log_dir, "black.log")
    write_file(log_file, output)
    emit_problem_matchers(log_file, "blackWarning.json", "black-warning")

    black_reformats, black_fails = parse_black_output(output)
    logging.info(
        "Found %s files would be reformatted and %s files would fail to reformat",
        black_reformats,
        black_fails,
    )

    report = generate_markdown_report(black_reformats, black_fails, log_file)
    append_file(args.report_file, report)

    sys.exit(0 if exit_code == 0 else 1)


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: LGPL-2.1-or-later

# *****************************************************************************
# *   Copyright (c) 2020 Rene Bartsch, B.Sc. Informatics <rene@bartschnet.de> *
# *                                                                           *
# *   This file is part of the FreeCAD CAx development system.                *
# *                                                                           *
# *   This program is free software; you can redistribute it and/or modify    *
# *   it under the terms of the GNU Lesser General Public License (LGPL)      *
# *   as published by the Free Software Foundation; either version 2 of       *
# *   the License, or (at your option) any later version.                     *
# *   for detail see the LICENCE text file.                                   *
# *                                                                           *
# *   FreeCAD is distributed in the hope that it will be useful,              *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
# *   GNU Lesser General Public License for more details.                     *
# *                                                                           *
# *   You should have received a copy of the GNU Library General Public       *
# *   License along with FreeCAD; if not, write to the Free Software          *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307    *
# *   USA                                                                     *
# *                                                                           *
# ****************************************************************************/
"""Postprocessor to output real GCode for Max Computer GmbH nccad9."""
import FreeCAD
import Path.Post.Utils as PostUtils
import PathScripts.PathUtils as PathUtils
import datetime


TOOLTIP = """
This is a postprocessor file for the Path workbench. It is used to take
a pseudo-G-code fragment output by a Path object and output real G-code
suitable for the Max Computer GmbH nccad9 Computer Numeric Control.

Supported features:

- 3-axis milling
- manual tool change with tool number as comment
- spindle speed as comment

!!! gCode files must use the suffix .knc !!!

import nccad_post
nccad_post.export([object], "/path/to/file.knc", "")
"""

MACHINE_NAME = """Max Computer GmbH nccad9 MCS/KOSY"""

# gCode for changing tools
# M01 <String> ; Displays <String> and waits for user interaction
TOOL_CHANGE = """G77      ; Move to release position
M10 O6.0 ; Stop spindle
M01 Insert tool TOOL
G76      ; Move to reference point to ensure correct coordinates after tool change
M10 O6.1 ; Start spindle"""


# gCode finishing the program
POSTAMBLE = """G77      ; Move to release position
M10 O6.0 ; Stop spindle"""


# gCode header with information about CAD-software, post-processor
# and date/time
if FreeCAD.ActiveDocument:
    cam_file = FreeCAD.ActiveDocument.FileName
else:
    cam_file = "<None>"

HEADER = """;Exported by FreeCAD
;Post Processor: {}
;CAM file: {}
;Output Time: {}
""".format(
    __name__, cam_file, str(datetime.datetime.now())
)


def export(objectslist, filename, argstring):
    """Export the list of objects into a filename.

    Parameters
    ----------
    objectslists: list
        List of objects.

    filename: str
        Name of the output file ending in `'.knc'`.
    """
    gcode = HEADER

    for obj in objectslist:
        for command in PathUtils.getPathWithPlacement(obj).Commands:
            # Manipulate tool change commands
            if "M6" == command.Name:
                gcode += TOOL_CHANGE.replace("TOOL", str(int(command.Parameters["T"])))
            elif "M3" == command.Name:
                # Convert spindle speed (rpm) command to comment
                gcode += (
                    "M01 Set spindle speed to "
                    + str(int(command.Parameters["S"]))
                    + " rounds per minute"
                )
            else:
                # Add other commands
                gcode += command.Name

                # Loop through command parameters
                for parameter, value in command.Parameters.items():
                    # Multiply F parameter value by 10,
                    # FreeCAD = mm/s, nccad = 1/10 mm/s
                    if "F" == parameter:
                        value *= 10
                    # Add command parameters and values and round float
                    # as nccad9 does not support exponents
                    gcode += " " + parameter + str(round(value, 5))

            gcode += "\n"

    gcode += POSTAMBLE + "\n"

    # Open editor window
    if FreeCAD.GuiUp:
        dia = PostUtils.GCodeEditorDialog()
        dia.editor.setText(gcode)
        result = dia.exec_()
        if result:
            gcode = dia.editor.toPlainText()

    # Save to file
    if filename != "-":
        gfile = open(filename, "w")
        gfile.write(gcode)
        gfile.close()

    return filename



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2017 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import math
from datetime import datetime

import FreeCAD
import DraftGeomUtils
import DraftVecUtils
import Part
import WorkingPlane

# This is roughly based on the no-fit polygon algorithm, used in
# SvgNest: https://github.com/Jack000/SVGnest
# Wikihouse plugin: https://github.com/tav/wikihouse-plugin/blob/master/wikihouse.rb

TOLERANCE = 0.0001  # smaller than this, two points are considered equal
DISCRETIZE = 4  # the number of segments in which arcs must be subdivided
ROTATIONS = [0, 90, 180, 270]  # the possible rotations to try


class Nester:

    def __init__(self, container=None, shapes=None):
        """Nester([container,shapes]): Creates a nester object with a container
        shape and a list of other shapes to nest into it. Container and
        shapes must be Part.Faces.

        Typical workflow:

        n = Nester() # creates the nester
        n.addContainer(object) # adds a doc object as the container
        n.addObjects(objects) # adds a list of doc objects as shapes
        n.run() # runs the nesting
        n.show() # creates a preview (compound) of the results
        n.apply() # applies transformations to the original objects

        Defaults (can be changed):

        Nester.TOLERANCE = 0.0001
        Nester.DISCRETIZE = 4
        Nester.ROTATIONS = [0,90,180,270]
        """

        self.objects = None
        self.container = container
        self.shapes = shapes
        self.results = []  # storage for the different results
        self.indexedFaces = None
        self.running = True
        self.progress = 0
        self.setCounter = None  # optionally define a setCounter(value) function where value is a %

    def addObjects(self, objects):
        """addObjects(objects): adds FreeCAD DocumentObjects to the nester"""

        if not isinstance(objects, list):
            objects = [objects]
        if not self.objects:
            self.objects = {}
        if not self.shapes:
            self.shapes = []
        for obj in objects:
            if hasattr(obj, "Shape"):
                h = obj.Shape.hashCode()
                if not h in self.objects:
                    self.objects[h] = obj
                    self.shapes.append(obj.Shape)

    def addContainer(self, container):
        """addContainer(object): adds a FreeCAD DocumentObject as the container"""

        if hasattr(container, "Shape"):
            self.container = container.Shape

    def clear(self):
        """clear(): Removes all objects and shape from the nester"""

        self.objects = None
        self.shapes = None

    def stop(self):
        """stop((): stops the computation"""

        self.running = False

    def update(self):
        """update(): internal function to verify if computation can
        go on"""

        if self.setCounter:
            self.setCounter(self.progress)
        if FreeCAD.GuiUp:
            from PySide import QtGui

            QtGui.QApplication.processEvents()
        if not self.running:
            return False
        return True

    def run(self):
        """run(): Runs a nesting operation. Returns a list of lists of
        shapes, each primary list being one filled container, or None
        if the operation failed."""

        # reset abort mechanism and variables

        self.running = True
        self.progress = 0
        starttime = datetime.now()

        # general conformity tests

        print("Executing conformity tests… ", end="")
        if not self.container:
            print("Empty container. Aborting.")
            return
        if not self.shapes:
            print("Empty shapes. Aborting.")
            return
        if not isinstance(self.container, Part.Face):
            print("Container is not a face. Aborting.")
            return
        normal = self.container.normalAt(0, 0)
        for s in self.shapes:
            if not self.update():
                return
            if len(s.Faces) != 1:
                print("One of the shapes does not contain exactly one face. Aborting.")
                return
            # check if all faces correctly oriented (same normal)
            if s.Faces[0].normalAt(0, 0).getAngle(normal) > TOLERANCE:
                # let pass faces with inverted normal
                if s.Faces[0].normalAt(0, 0).getAngle(normal) - math.pi > TOLERANCE:
                    print(
                        "One of the face does not have the same orientation as the container. Aborting."
                    )
                    return

        # TODO
        # allow one to use a non-rectangular container
        # manage margins/paddings
        # allow one to prevent or force specific rotations for a piece

        # LONG-TERM TODO
        # add genetic algo to swap pieces, and check if the result is better

        # track progresses
        step = 100.0 / (len(self.shapes) * len(ROTATIONS))

        # store hashCode together with the face so we can change the order
        # and still identify the original face, so we can calculate a transform afterwards
        self.indexedfaces = [[shape.hashCode(), shape] for shape in self.shapes]

        # build a clean copy so we don't touch the original
        faces = list(self.indexedfaces)

        # replace shapes by their face
        faces = [[f[0], f[1].Faces[0]] for f in faces]

        # order by area
        faces = sorted(faces, key=lambda face: face[1].Area)

        # discretize non-linear edges and remove holes
        nfaces = []
        for face in faces:
            if not self.update():
                return
            nedges = []
            allLines = True
            for edge in face[1].OuterWire.OrderedEdges:
                if isinstance(edge.Curve, (Part.LineSegment, Part.Line)):
                    nedges.append(edge)
                else:
                    allLines = False
                    last = edge.Vertexes[0].Point
                    for i in range(DISCRETIZE):
                        s = float(i + 1) / DISCRETIZE
                        par = edge.FirstParameter + (edge.LastParameter - edge.FirstParameter) * s
                        new = edge.valueAt(par)
                        nedges.append(Part.LineSegment(last, new).toShape())
                        last = new
            f = Part.Face(Part.Wire(nedges))
            if not f.isValid():
                if allLines:
                    print("Invalid face found in set. Aborting.")
                else:
                    print("Face distretizing failed. Aborting.")
                return
            nfaces.append([face[0], f])
        faces = nfaces

        # container for sheets with a first, empty sheet
        sheets = [[]]

        print("Everything OK (", datetime.now() - starttime, ")")

        # main loop

        facenumber = 1
        facesnumber = len(faces)

        # print("Vertices per face:",[len(face[1].Vertexes) for face in faces])

        while faces:

            print(
                "Placing piece",
                facenumber,
                "/",
                facesnumber,
                "Area:",
                FreeCAD.Units.Quantity(faces[-1][1].Area, FreeCAD.Units.Area).getUserPreferred()[0],
                ": ",
                end="",
            )

            face = faces.pop()
            boc = self.container.BoundBox

            # this stores the available solutions for each rotation of a piece
            # contains [sheetnumber,face,xlength] lists,
            # face being [hascode,transformed face] and xlength
            # the X size of all boundboxes of placed pieces
            available = []

            # this stores the possible positions on a blank
            # sheet, in case we need to create a new one
            initials = []

            # this checks if the piece don't fit in the container
            unfit = True

            for rotation in ROTATIONS:

                if not self.update():
                    return

                self.progress += step

                print(rotation, ", ", end="")
                hashcode = face[0]
                rotface = face[1].copy()
                if rotation:
                    rotface.rotate(rotface.CenterOfMass, normal, rotation)
                bof = rotface.BoundBox
                rotverts = self.order(rotface)
                # for i,v in enumerate(rotverts):
                #    Draft.makeText([str(i)],point=v)
                basepoint = rotverts[0]  # leftmost point of the rotated face
                basecorner = boc.getPoint(0)  # lower left corner of the container

                # See if the piece fits in the container dimensions
                if (bof.XLength < boc.XLength) and (bof.YLength < boc.YLength):
                    unfit = False

                # Get the fit polygon of the container
                # that is, the polygon inside which basepoint can
                # circulate, and the face still be fully inside the container

                v1 = basecorner.add(basepoint.sub(bof.getPoint(0)))
                v2 = v1.add(FreeCAD.Vector(0, boc.YLength - bof.YLength, 0))
                v3 = v2.add(FreeCAD.Vector(boc.XLength - bof.XLength, 0, 0))
                v4 = v3.add(FreeCAD.Vector(0, -(boc.YLength - bof.YLength), 0))
                binpol = Part.Face(Part.makePolygon([v1, v2, v3, v4, v1]))
                initials.append([binpol, [hashcode, rotface], basepoint])

                # check for available space on each existing sheet

                for sheetnumber, sheet in enumerate(sheets):
                    # Get the no-fit polygon for each already placed face in
                    # current sheet. That is, a polygon in which basepoint
                    # cannot be, if we want our face to not overlap with the
                    # placed face.
                    # To do this, we "circulate" the face around the placed face

                    if not self.update():
                        return

                    nofitpol = []
                    for placed in sheet:
                        pts = []
                        for placedvert in self.order(placed[1], right=True):
                            fpts = []
                            for i, rotvert in enumerate(rotverts):
                                if not self.update():
                                    return

                                facecopy = rotface.copy()
                                facecopy.translate(placedvert.sub(rotvert))

                                # test if all the points of the face are outside the
                                # placed face (except the base point, which is coincident)

                                outside = True
                                faceverts = self.order(facecopy)
                                for vert in faceverts:
                                    if (vert.sub(placedvert)).Length > TOLERANCE:
                                        if placed[1].isInside(vert, TOLERANCE, True):
                                            outside = False
                                            break

                                # also need to test for edge intersection, because even
                                # if all vertices are outside, the pieces could still
                                # overlap

                                if outside:
                                    for e1 in facecopy.OuterWire.Edges:
                                        for e2 in placed[1].OuterWire.Edges:
                                            if not self.update():
                                                return

                                            if True:
                                                # Draft code (SLOW)
                                                p = DraftGeomUtils.findIntersection(e1, e2)
                                                if p:
                                                    p = p[0]
                                                    p1 = e1.Vertexes[0].Point
                                                    p2 = e1.Vertexes[1].Point
                                                    p3 = e2.Vertexes[0].Point
                                                    p4 = e2.Vertexes[1].Point
                                                    if (
                                                        (p.sub(p1).Length > TOLERANCE)
                                                        and (p.sub(p2).Length > TOLERANCE)
                                                        and (p.sub(p3).Length > TOLERANCE)
                                                        and (p.sub(p4).Length > TOLERANCE)
                                                    ):
                                                        outside = False
                                                        break
                                            else:
                                                # alt code: using distToShape (EVEN SLOWER!)
                                                p = e1.distToShape(e2)
                                                if p:
                                                    if p[0] < TOLERANCE:
                                                        # allow vertex-to-vertex intersection
                                                        if (p[2][0][0] != "Vertex") or (
                                                            p[2][0][3] != "Vertex"
                                                        ):
                                                            outside = False
                                                            break

                                if outside:
                                    fpts.append([faceverts[0], i])
                                    # Draft.makeText([str(i)],point=faceverts[0])

                            # reorder available solutions around a same point if needed
                            # ensure they are in the correct order

                            idxs = [p[1] for p in fpts]
                            if (0 in idxs) and (len(faceverts) - 1 in idxs):
                                slicepoint = len(fpts)
                                last = len(faceverts)
                                for p in reversed(fpts):
                                    if p[1] == last - 1:
                                        slicepoint -= 1
                                        last -= 1
                                    else:
                                        break
                                fpts = fpts[slicepoint:] + fpts[:slicepoint]
                                # print(fpts)
                            pts.extend(fpts)

                        # create the polygon

                        if len(pts) < 3:
                            print("Error calculating a no-fit polygon. Aborting.")
                            return
                        pts = [p[0] for p in pts]
                        pol = Part.Face(Part.makePolygon(pts + [pts[0]]))

                        if not pol.isValid():

                            # fix overlapping edges

                            overlap = True
                            while overlap:
                                overlap = False
                                for i in range(len(pol.OuterWire.Edges) - 1):
                                    if not self.update():
                                        return

                                    v1 = DraftGeomUtils.vec(pol.OuterWire.OrderedEdges[i])
                                    v2 = DraftGeomUtils.vec(pol.OuterWire.OrderedEdges[i + 1])
                                    if abs(v1.getAngle(v2) - math.pi) <= TOLERANCE:
                                        overlap = True
                                        ne = Part.LineSegment(
                                            pol.OuterWire.OrderedEdges[i].Vertexes[0].Point,
                                            pol.OuterWire.OrderedEdges[i + 1].Vertexes[-1].Point,
                                        ).toShape()
                                        pol = Part.Face(
                                            Part.Wire(
                                                pol.OuterWire.OrderedEdges[:i]
                                                + [ne]
                                                + pol.OuterWire.OrderedEdges[i + 2 :]
                                            )
                                        )
                                        break

                        if not pol.isValid():

                            # trying basic OCC fix

                            pol.fix(0, 0, 0)
                            if pol.isValid():
                                if pol.ShapeType == "Face":
                                    pol = Part.Face(pol.OuterWire)  # discard possible inner holes
                                elif pol.Faces:
                                    # several faces after the fix, keep the biggest one
                                    a = 0
                                    ff = None
                                    for f in pol.Faces:
                                        if f.Area > a:
                                            a = f.Area
                                            ff = f
                                    if ff:
                                        pol = ff
                                else:
                                    print("Unable to fix invalid no-fit polygon. Aborting.")
                                    Part.show(pol)
                                    return

                        if not pol.isValid():

                            # none of the fixes worked. Epic fail.

                            print("Invalid no-fit polygon. Aborting.")
                            Part.show(pol.OuterWire)
                            for p in sheet:
                                Part.show(p[1])
                            Part.show(facecopy)
                            # for i,p in enumerate(faceverts):
                            #    Draft.makeText([str(i)],point=p)
                            return

                        if pol.isValid():
                            nofitpol.append(pol)
                            # Part.show(pol)

                    # Union all the no-fit pols into one

                    if len(nofitpol) == 1:
                        nofitpol = nofitpol[0]
                    elif len(nofitpol) > 1:
                        b = nofitpol.pop()
                        for n in nofitpol:
                            if not self.update():
                                return
                            b = b.fuse(n)
                        nofitpol = b

                        # remove internal edges (discard edges shared by 2 faces)

                        lut = {}
                        for f in fitpol.Faces:
                            for e in f.Edges:
                                h = e.hashCode()
                                if h in lut:
                                    lut[h].append(e)
                                else:
                                    lut[h] = [e]
                        edges = [e[0] for e in lut.values() if len(e) == 1]
                        try:
                            pol = Part.Face(Part.Wire(edges))
                        except Exception:
                            # above method can fail sometimes. Try a slower method
                            w = DraftGeomUtils.findWires(edges)
                            if len(w) == 1:
                                if w[0].isClosed():
                                    try:
                                        pol = Part.Face(w[0])
                                    except Exception:
                                        print("Error merging polygons. Aborting.")
                                        try:
                                            Part.show(Part.Wire(edges))
                                        except Exception:
                                            for e in edges:
                                                Part.show(e)
                                        return

                    # subtract the no-fit polygon from the container's fit polygon
                    # we then have the zone where the face can be placed

                    if nofitpol:
                        fitpol = binpol.cut(nofitpol)
                    else:
                        fitpol = binpol.copy()

                    # check that we have some space on this sheet

                    if (fitpol.Area > 0) and fitpol.Vertexes:

                        # order the fitpol vertexes by smallest X
                        # and try to place the piece, making sure it doesn't
                        # intersect with already placed pieces
                        fitverts = sorted([v.Point for v in fitpol.Vertexes], key=lambda v: v.x)
                        for p in fitverts:
                            if not self.update():
                                return

                            trface = rotface.copy()
                            trface.translate(p.sub(basepoint))
                            ok = True
                            for placed in sheet:
                                if ok:
                                    for vert in trface.Vertexes:
                                        if placed[1].isInside(vert.Point, TOLERANCE, False):
                                            ok = False
                                            break
                                if ok:
                                    for e1 in trface.OuterWire.Edges:
                                        for e2 in placed[1].OuterWire.Edges:
                                            p = DraftGeomUtils.findIntersection(e1, e2)
                                            if p:
                                                p = p[0]
                                                p1 = e1.Vertexes[0].Point
                                                p2 = e1.Vertexes[1].Point
                                                p3 = e2.Vertexes[0].Point
                                                p4 = e2.Vertexes[1].Point
                                                if (
                                                    (p.sub(p1).Length > TOLERANCE)
                                                    and (p.sub(p2).Length > TOLERANCE)
                                                    and (p.sub(p3).Length > TOLERANCE)
                                                    and (p.sub(p4).Length > TOLERANCE)
                                                ):
                                                    ok = False
                                                    break
                                if not ok:
                                    break
                            if ok:
                                rotface = trface
                                break
                        else:
                            print("Could not determine location on sheet. Aborting.")
                            return

                        # check the X space occupied by this solution

                        bb = rotface.BoundBox
                        for placed in sheet:
                            bb.add(placed[1].BoundBox)
                        available.append([sheetnumber, [hashcode, rotface], bb.XMax, fitpol])

            if unfit:
                print("One face does not fit in the container. Aborting.")
                return

            if available:

                # order by smallest X size and take the first one
                available = sorted(available, key=lambda sol: sol[2])
                print("Adding piece to sheet", available[0][0] + 1)
                sheets[available[0][0]].append(available[0][1])
                # Part.show(available[0][3])

            else:

                # adding to the leftmost vertex of the binpol

                sheet = []
                print("Creating new sheet, adding piece to sheet", len(sheets))
                # order initial positions by smallest X size
                initials = sorted(initials, key=lambda sol: sol[1][1].BoundBox.XLength)
                hashcode = initials[0][1][0]
                face = initials[0][1][1]
                # order binpol vertexes by X coord
                verts = sorted([v.Point for v in initials[0][0].Vertexes], key=lambda v: v.x)
                face.translate(verts[0].sub(initials[0][2]))
                sheet.append([hashcode, face])
                sheets.append(sheet)

            facenumber += 1

        print("Run time:", datetime.now() - starttime)
        self.results.append(sheets)
        return sheets

    def order(self, face, right=False):
        """order(face,[right]): returns a list of vertices
        ordered clockwise. The first vertex will be the
        lefmost one, unless right is True, in which case the
        first vertex will be the rightmost one"""

        verts = [v.Point for v in face.OuterWire.OrderedVertexes]

        # flatten the polygon on the XY plane

        wp = WorkingPlane.PlaneBase()
        wp.align_to_point_and_axis(face.CenterOfMass, face.normalAt(0, 0))
        pverts = []
        for v in verts:
            vx = DraftVecUtils.project(v, wp.u)
            lx = vx.Length
            if vx.getAngle(wp.u) > 1:
                lx = -lx
            vy = DraftVecUtils.project(v, wp.v)
            ly = vy.Length
            if vy.getAngle(wp.v) > 1:
                ly = -ly
            pverts.append(FreeCAD.Vector(lx, ly, 0))
        pverts.append(pverts[0])

        # https://stackoverflow.com/questions/1165647/how-to-determine-if-a-list-of-polygon-points-are-in-clockwise-order

        s = 0
        for i in range(len(pverts) - 1):
            s += (pverts[i + 1].x - pverts[i].x) * (pverts[i + 1].y + pverts[i].y)
        if s < 0:
            verts.reverse()
        elif s == 0:
            print("error computing winding direction")
            return

        return verts

    def show(self, result=None):
        """show([result]): creates shapes in the document, showing
        the given result (list of sheets) or the last result if
        none is provided"""

        if not result:
            result = []
            if self.results:
                result = self.results[-1]
        offset = FreeCAD.Vector(0, 0, 0)
        feats = []
        for sheet in result:
            shapes = [self.container.OuterWire]
            shapes.extend([face[1] for face in sheet])
            comp = Part.makeCompound(shapes)
            comp.translate(offset)
            o = FreeCAD.ActiveDocument.addObject("Part::Feature", "Nest")
            o.Shape = comp
            feats.append(o)
            offset = offset.add(FreeCAD.Vector(1.1 * self.container.BoundBox.XLength, 0, 0))
        FreeCAD.ActiveDocument.recompute()
        return feats

    def getPlacements(self, result=None):
        """getPlacements([result]): returns a dictionary of hashCode:Placement
        pairs from the given result or the last computed result if none
        is given. The Placement contains a translation vector and a rotation
        to be given to the final object."""

        if not self.indexedfaces:
            print("error: shapes were not indexed. Use run() first")
            return
        if not result:
            result = []
            if self.results:
                result = self.results[-1]
        d = {}
        offset = FreeCAD.Vector(0, 0, 0)
        for sheet in result:
            for face in sheet:
                orig = None
                for pair in self.indexedfaces:
                    if pair[0] == face[0]:
                        orig = pair[1]
                if not orig:
                    print("error: hashCode mismatch between original and transformed face")
                    return
                shape = face[1]
                if offset.Length:
                    shape.translate(offset)
                deltav = shape.Faces[0].CenterOfMass.sub(orig.Faces[0].CenterOfMass)
                rot = FreeCAD.Rotation(
                    orig.Vertexes[0].Point.sub(orig.Faces[0].CenterOfMass),
                    shape.Vertexes[0].Point.sub(shape.Faces[0].CenterOfMass),
                )
                pla = FreeCAD.Placement(deltav, rot)
                d[face[0]] = pla
            offset = offset.add(FreeCAD.Vector(1.1 * self.container.BoundBox.XLength, 0, 0))
        return d

    def apply(self, result=None):
        """apply([result]): Applies the computed placements of the given
        result, or the last computed result if none is given, to the
        document objects given to the nester via addObjects() before
        running."""

        if not self.objects:
            print("objects list is empty")
            return
        p = self.getPlacements(result)
        if p:
            for key, pla in p.items():
                if key in self.objects:
                    sh = self.objects[key].Shape.copy()
                    sh.translate(pla.Base)
                    sh.rotate(
                        sh.Faces[0].CenterOfMass,
                        pla.Rotation.Axis,
                        math.degrees(pla.Rotation.Angle),
                    )
                    self.objects[key].Placement = sh.Placement
                else:
                    print("error: hashCode mismatch with original object")


def test():
    "runs a test with selected shapes, container selected last"

    import FreeCADGui

    sel = FreeCADGui.Selection.getSelection()
    if sel:
        container = sel.pop().Shape
        shapes = [o.Shape for o in sel]
        n = Nester(container, shapes)
        result = n.run()
        if result:
            n.show()



from Base.Metadata import export, constmethod
from Base.Persistence import Persistence
from typing import Any, Final, Union, List, Optional


@export(
    DisableNotify=True,
)
class PropertyContainer(Persistence):
    """
    App.PropertyContainer class.
    """

    PropertiesList: Final[list] = []
    """A list of all property names."""

    def getPropertyByName(self, name: str, checkOwner: int = 0) -> Any:
        """
        getPropertyByName(name, checkOwner=0) -> object or Tuple

        Returns the value of a named property. Note that the returned property may not
        always belong to this container (e.g. from a linked object).

        name : str
            Name of the property.
        checkOwner : int
            0: just return the property.
            1: raise exception if not found or the property does not belong to this container.
            2: return a tuple (owner, propertyValue).
        """
        ...

    def getPropertyTouchList(self, name: str) -> tuple:
        """
        getPropertyTouchList(name) -> tuple

        Returns a list of index of touched values for list type properties.

        name : str
            Property name.
        """
        ...

    def getTypeOfProperty(self, name: str) -> list:
        """
        getTypeOfProperty(name) -> list

        Returns the type of a named property. This can be a list conformed by elements in
        (Hidden, NoRecompute, NoPersist, Output, ReadOnly, Transient).

        name : str
            Property name.
        """
        ...

    def getTypeIdOfProperty(self, name: str) -> str:
        """
        getTypeIdOfProperty(name) -> str

        Returns the C++ class name of a named property.

        name : str
            Property name.
        """
        ...

    def setEditorMode(self, name: str, type: Union[int, List[str]]) -> None:
        """
        setEditorMode(name, type) -> None

        Set the behaviour of the property in the property editor.

        name : str
            Property name.
        type : int, sequence of str
            Property type.
            0: default behaviour. 1: item is ready-only. 2: item is hidden. 3: item is hidden and read-only.
            If sequence, the available items are 'ReadOnly' and 'Hidden'.
        """
        ...

    def getEditorMode(self, name: str) -> list:
        """
        getEditorMode(name) -> list

        Get the behaviour of the property in the property editor.
        It returns a list of strings with the current mode. If the list is empty there are no
        special restrictions.
        If the list contains 'ReadOnly' then the item appears in the property editor but is
        disabled.
        If the list contains 'Hidden' then the item even doesn't appear in the property editor.

        name : str
            Property name.
        """
        ...

    def getGroupOfProperty(self, name: str) -> str:
        """
        getGroupOfProperty(name) -> str

        Returns the name of the group which the property belongs to in this class.
        The properties are sorted in different named groups for convenience.

        name : str
            Property name.
        """
        ...

    def setGroupOfProperty(self, name: str, group: str) -> None:
        """
        setGroupOfProperty(name, group) -> None

        Set the name of the group of a dynamic property.

        name : str
            Property name.
        group : str
            Group name.
        """
        ...

    def setPropertyStatus(
        self, name: str, val: Union[int, str, List[Union[str, int]]]
    ) -> None:
        """
        setPropertyStatus(name, val) -> None

        Set property status.

        name : str
            Property name.
        val : int, str, sequence of str or int
            Call getPropertyStatus() to get a list of supported text value.
            If the text start with '-' or the integer value is negative, then the status is cleared.
        """
        ...

    def getPropertyStatus(self, name: str = "") -> list:
        """
        getPropertyStatus(name='') -> list

        Get property status.

        name : str
            Property name. If empty, returns a list of supported text names of the status.
        """
        ...

    def getDocumentationOfProperty(self, name: str) -> str:
        """
        getDocumentationOfProperty(name) -> str

        Returns the documentation string of the property of this class.

        name : str
            Property name.
        """
        ...

    def setDocumentationOfProperty(self, name: str, docstring: str) -> None:
        """
        setDocumentationOfProperty(name, docstring) -> None

        Set the documentation string of a dynamic property of this class.

        name : str
            Property name.
        docstring : str
            Documentation string.
        """
        ...

    def getEnumerationsOfProperty(self, name: str) -> Optional[list]:
        """
        getEnumerationsOfProperty(name) -> list or  None

        Return all enumeration strings of the property of this class or None if not a
        PropertyEnumeration.

        name : str
            Property name.
        """
        ...

    @constmethod
    def dumpPropertyContent(self, Property: str, *, Compression: int = 3) -> bytearray:
        """
        dumpPropertyContent(Property, Compression=3) -> bytearray

        Dumps the content of the property, both the XML representation and the additional
        data files required, into a byte representation.

        Property : str
            Property Name.
        Compression : int
            Set the data compression level in the range [0, 9]. Set to 0 for no compression.
        """
        ...

    def restorePropertyContent(self, name: str, obj: object) -> None:
        """
        restorePropertyContent(name, obj) -> None

        Restore the content of the object from a byte representation as stored by `dumpPropertyContent`.
        It could be restored from any Python object implementing the buffer protocol.

        name : str
            Property name.
        obj : buffer
            Object with buffer protocol support.
        """
        ...

    @constmethod
    def renameProperty(self, oldName: str, newName: str) -> None:
        """
        renameProperty(oldName, newName) -> None

        Rename a property.

        oldName : str
            Old property name.
        newName : str
            New property name.
        """
        ...



# SPDX-License-Identifier: LGPL-2.1-or-later

#**************************************************************************
#   Copyright (c) 2023 David Carter <dcarter@davidcarter.ca>              *
#                                                                         *
#   This file is part of the FreeCAD CAx development system.              *
#                                                                         *
#   This program is free software; you can redistribute it and/or modify  *
#   it under the terms of the GNU Lesser General Public License (LGPL)    *
#   as published by the Free Software Foundation; either version 2 of     *
#   the License, or (at your option) any later version.                   *
#   for detail see the LICENCE text file.                                 *
#                                                                         *
#   FreeCAD is distributed in the hope that it will be useful,            *
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#   GNU Library General Public License for more details.                  *
#                                                                         *
#   You should have received a copy of the GNU Library General Public     *
#   License along with FreeCAD; if not, write to the Free Software        *
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#   USA                                                                   *
#**************************************************************************

"""
Test module for FreeCAD material models
"""

import unittest
import FreeCAD
import Materials

parseQuantity = FreeCAD.Units.parseQuantity

class ModelTestCases(unittest.TestCase):
    """
    Test class for FreeCAD material models
    """
    def setUp(self):
        """ Setup function to initialize test data """
        self.ModelManager = Materials.ModelManager()
        self.uuids = Materials.UUIDs()

    def testModelManager(self):
        """ Ensure we can access ModelManager member functions """
        self.assertIn("ModelLibraries", dir(self.ModelManager))
        self.assertIn("Models", dir(self.ModelManager))

    def testUUIDs(self):
        """ Verify the common UUIDs are defined and correct """
        self.assertTrue(self.uuids.Father, "9cdda8b6-b606-4778-8f13-3934d8668e67")
        self.assertTrue(self.uuids.MaterialStandard, "1e2c0088-904a-4537-925f-64064c07d700")

        self.assertTrue(self.uuids.ArrudaBoyce, "e10d00de-c7de-4e59-bcdd-058c2ea19ec6")
        self.assertTrue(self.uuids.Density, "454661e5-265b-4320-8e6f-fcf6223ac3af")
        self.assertTrue(self.uuids.Hardness, "3d1a6141-d032-4d82-8bb5-a8f339fff8ad")
        self.assertTrue(self.uuids.IsotropicLinearElastic, "f6f9e48c-b116-4e82-ad7f-3659a9219c50")
        self.assertTrue(self.uuids.LinearElastic,"7b561d1d-fb9b-44f6-9da9-56a4f74d7536")
        self.assertTrue(self.uuids.MooneyRivlin, "beeed169-7770-4da0-ab67-c9172cf7d23d")
        self.assertTrue(self.uuids.NeoHooke, "569ebc58-ef29-434a-83be-555a0980d505")
        self.assertTrue(self.uuids.OgdenN1, "a2634a2c-412f-468d-9bec-74ae5d87a9c0")
        self.assertTrue(self.uuids.OgdenN2, "233540bb-7b13-4f49-ac12-126a5c82cedf")
        self.assertTrue(self.uuids.OgdenN3, "a917d6b8-209f-429e-9972-fe4bbb97af3f")
        self.assertTrue(self.uuids.OgdenYld2004p18, "3ef9e427-cc25-43f7-817f-79ff0d49625f")
        self.assertTrue(self.uuids.OrthotropicLinearElastic, "b19ccc6b-a431-418e-91c2-0ac8c649d146")
        self.assertTrue(self.uuids.PolynomialN1, "285a6042-0f0c-4a36-a898-4afadd6408ce")
        self.assertTrue(self.uuids.PolynomialN2, "4c2fb7b2-5121-4d6f-be0d-8c5970c9e682")
        self.assertTrue(self.uuids.PolynomialN3, "e83ada22-947e-4beb-91e7-482a16f5ba77")
        self.assertTrue(self.uuids.ReducedPolynomialN1, "f8052a3c-db17-42ea-b2be-13aa5ef30730")
        self.assertTrue(self.uuids.ReducedPolynomialN2, "c52b5021-4bb8-441c-80d4-855fce9de15e")
        self.assertTrue(self.uuids.ReducedPolynomialN3, "fa4e58b4-74c7-4292-8e79-7d5fd232fb55")
        self.assertTrue(self.uuids.Yeoh, "cd13c492-21a9-4578-8191-deec003e4c01")

        self.assertTrue(self.uuids.Fluid, "1ae66d8c-1ba1-4211-ad12-b9917573b202")

        self.assertTrue(self.uuids.Thermal, "9959d007-a970-4ea7-bae4-3eb1b8b883c7")

        self.assertTrue(self.uuids.Electromagnetic, "b2eb5f48-74b3-4193-9fbb-948674f427f3")

        self.assertTrue(self.uuids.Architectural, "32439c3b-262f-4b7b-99a8-f7f44e5894c8")
        self.assertTrue(self.uuids.ArchitecturalRendering, "27e48ac9-54e1-4a1f-aa49-d5d690242705")

        self.assertTrue(self.uuids.Costs, "881df808-8726-4c2e-be38-688bb6cce466")

        self.assertTrue(self.uuids.BasicRendering, "f006c7e4-35b7-43d5-bbf9-c5d572309e6e")
        self.assertTrue(self.uuids.TextureRendering, "bbdcc65b-67ca-489c-bd5c-a36e33d1c160")
        self.assertTrue(self.uuids.AdvancedRendering, "c880f092-cdae-43d6-a24b-55e884aacbbf")
        self.assertTrue(self.uuids.VectorRendering, "fdf5a80e-de50-4157-b2e5-b6e5f88b680e")

        self.assertTrue(self.uuids.RenderAppleseed, "b0a10f70-13bf-4598-ab63-bcfbbcd813e3")
        self.assertTrue(self.uuids.RenderCarpaint, "4d2cc163-0707-40e2-a9f7-14288c4b97bd")
        self.assertTrue(self.uuids.RenderCycles, "a6da1b66-929c-48bf-ae80-3b0495c7b50b")
        self.assertTrue(self.uuids.RenderDiffuse, "c19b2d30-c55b-48aa-a938-df9e2f7779cf")
        self.assertTrue(self.uuids.RenderDisney, "f8723572-4470-4c39-a749-6d3b71358a5b")
        self.assertTrue(self.uuids.RenderEmission, "9f6cb588-c89d-4a74-9d0f-2786a8568cec")
        self.assertTrue(self.uuids.RenderGlass, "d76a56f5-7250-4efb-bb89-8ea0a9ccaa6b")
        self.assertTrue(self.uuids.RenderLuxcore, "6b992304-33e0-490b-a391-e9d0af79bb69")
        self.assertTrue(self.uuids.RenderLuxrender, "67ac6a63-e173-4e05-898b-af743f1f9563")
        self.assertTrue(self.uuids.RenderMixed, "84bab333-984f-47fe-a512-d17c7cb2daa9")
        self.assertTrue(self.uuids.RenderOspray, "a4792c23-0be9-47c2-b16d-47b2d2d5efd6")
        self.assertTrue(self.uuids.RenderPbrt, "35b34b82-4325-4d27-97bd-d10bb2c56586")
        self.assertTrue(self.uuids.RenderPovray, "6ec8b415-4c7b-4206-a80b-2ea64101f34b")
        self.assertTrue(self.uuids.RenderSubstancePBR, "f212b643-db96-452e-8428-376a4534e5ab")
        self.assertTrue(self.uuids.RenderTexture, "fc9b6135-95cd-4ba8-ad9a-0972caeebad2")
        self.assertTrue(self.uuids.RenderWB, "344008be-a837-43af-90bc-f795f277b309")

        self.assertTrue(self.uuids.TestModel, "34d0583d-f999-49ba-99e6-aa40bd5c3a6b")

    def testModelLoad(self):
        """ Test that the Density model has been loaded correctly """
        density = self.ModelManager.getModel(self.uuids.Density)
        self.assertIsNotNone(density)
        self.assertEqual(density.Name, "Density")
        self.assertEqual(density.UUID, "454661e5-265b-4320-8e6f-fcf6223ac3af")
        self.assertIn("Density", density.Properties)
        prop = density.Properties["Density"]
        self.assertIn("Description", dir(prop))
        self.assertIn("Name", dir(prop))
        self.assertIn("Type", dir(prop))
        self.assertIn("URL", dir(prop))
        self.assertIn("Units", dir(prop))
        self.assertEqual(prop.Name, "Density")

    def testTestModelCompleteness(self):
        """ Test that the Test model has been loaded correctly """
        model = self.ModelManager.getModel(self.uuids.TestModel)
        self.assertIsNotNone(model)
        self.assertEqual(model.Name, "Test Model")
        self.assertEqual(model.UUID, "34d0583d-f999-49ba-99e6-aa40bd5c3a6b")
        self.assertIn("TestString", model.Properties)
        self.assertEqual(len(model.Properties), 17)
        prop = model.Properties["TestString"]
        self.assertIn("Description", dir(prop))
        self.assertIn("Name", dir(prop))
        self.assertIn("Type", dir(prop))
        self.assertIn("URL", dir(prop))
        self.assertIn("Units", dir(prop))
        self.assertEqual(prop.Name, "TestString")
        self.assertEqual(prop.Type, "String")
        self.assertEqual(prop.URL, "")
        self.assertEqual(prop.Units, "")
        self.assertEqual(prop.Description, "A String")
        prop = model.Properties["TestURL"]
        self.assertEqual(prop.Name, "TestURL")
        self.assertEqual(prop.Type, "URL")
        self.assertEqual(prop.URL, "")
        self.assertEqual(prop.Units, "")
        self.assertEqual(prop.Description, "A URL")
        prop = model.Properties["TestList"]
        self.assertEqual(prop.Name, "TestList")
        self.assertEqual(prop.Type, "List")
        self.assertEqual(prop.URL, "")
        self.assertEqual(prop.Units, "")
        self.assertEqual(prop.Description, "A List")
        prop = model.Properties["TestFileList"]
        self.assertEqual(prop.Name, "TestFileList")
        self.assertEqual(prop.Type, "FileList")
        self.assertEqual(prop.URL, "")
        self.assertEqual(prop.Units, "")
        self.assertEqual(prop.Description, "A List of file paths")
        prop = model.Properties["TestImageList"]
        self.assertEqual(prop.Name, "TestImageList")
        self.assertEqual(prop.Type, "ImageList")
        self.assertEqual(prop.URL, "")
        self.assertEqual(prop.Units, "")
        self.assertEqual(prop.Description, "A List of embedded images")
        prop = model.Properties["TestInteger"]
        self.assertEqual(prop.Name, "TestInteger")
        self.assertEqual(prop.Type, "Integer")
        self.assertEqual(prop.URL, "")
        self.assertEqual(prop.Units, "")
        self.assertEqual(prop.Description, "A Integer")
        prop = model.Properties["TestFloat"]
        self.assertEqual(prop.Name, "TestFloat")
        self.assertEqual(prop.Type, "Float")
        self.assertEqual(prop.URL, "")
        self.assertEqual(prop.Units, "")
        self.assertEqual(prop.Description, "A Float")
        prop = model.Properties["TestBoolean"]
        self.assertEqual(prop.Name, "TestBoolean")
        self.assertEqual(prop.Type, "Boolean")
        self.assertEqual(prop.URL, "")
        self.assertEqual(prop.Units, "")
        self.assertEqual(prop.Description, "A Boolean")
        prop = model.Properties["TestColor"]
        self.assertEqual(prop.Name, "TestColor")
        self.assertEqual(prop.Type, "Color")
        self.assertEqual(prop.URL, "")
        self.assertEqual(prop.Units, "")
        self.assertEqual(prop.Description, "A Color")
        prop = model.Properties["TestFile"]
        self.assertEqual(prop.Name, "TestFile")
        self.assertEqual(prop.Type, "File")
        self.assertEqual(prop.URL, "")
        self.assertEqual(prop.Units, "")
        self.assertEqual(prop.Description, "A File")
        prop = model.Properties["TestSVG"]
        self.assertEqual(prop.Name, "TestSVG")
        self.assertEqual(prop.Type, "SVG")
        self.assertEqual(prop.URL, "")
        self.assertEqual(prop.Units, "")
        self.assertEqual(prop.Description, "An SVG")
        prop = model.Properties["TestImage"]
        self.assertEqual(prop.Name, "TestImage")
        self.assertEqual(prop.Type, "Image")
        self.assertEqual(prop.URL, "")
        self.assertEqual(prop.Units, "")
        self.assertEqual(prop.Description, "An Image")
        prop = model.Properties["TestQuantity"]
        self.assertEqual(prop.Name, "TestQuantity")
        self.assertEqual(prop.Type, "Quantity")
        self.assertEqual(prop.URL, "")
        self.assertEqual(prop.Units, "kg/m^3")
        self.assertEqual(prop.Description, "A Quantity")
        prop = model.Properties["TestMultiLineString"]
        self.assertEqual(prop.Name, "TestMultiLineString")
        self.assertEqual(prop.Type, "MultiLineString")
        self.assertEqual(prop.URL, "")
        self.assertEqual(prop.Units, "")
        self.assertEqual(prop.Description, "A string that spans multiple lines")

        prop = model.Properties["TestArray2D"]
        self.assertEqual(prop.Name, "TestArray2D")
        self.assertEqual(prop.Type, "2DArray")
        self.assertEqual(prop.URL, "")
        self.assertEqual(prop.Units, "")
        self.assertEqual(prop.Description, "2 Dimensional array showing density with temperature\n")
        self.assertEqual(len(prop.Columns), 2)
        col = prop.Columns[0]
        self.assertIn("Description", dir(col))
        self.assertIn("Name", dir(col))
        self.assertIn("Type", dir(col))
        self.assertIn("URL", dir(col))
        self.assertIn("Units", dir(col))
        self.assertEqual(col.Name, "Temperature")
        self.assertEqual(col.Type, "Quantity")
        self.assertEqual(col.URL, "")
        self.assertEqual(col.Units, "C")
        self.assertEqual(col.Description, "Temperature")
        col = prop.Columns[1]
        self.assertEqual(col.Name, "Density")
        self.assertEqual(col.Type, "Quantity")
        self.assertEqual(col.URL, "https://en.wikipedia.org/wiki/Density")
        self.assertEqual(col.Units, "kg/m^3")
        self.assertEqual(col.Description, "Density in [FreeCAD Density unit]")

        prop = model.Properties["TestArray2D3Column"]
        self.assertEqual(prop.Name, "TestArray2D3Column")
        self.assertEqual(prop.Type, "2DArray")
        self.assertEqual(prop.URL, "")
        self.assertEqual(prop.Units, "")
        self.assertEqual(prop.Description, "2 Dimensional array showing density and initial yield stress with temperature\n")
        self.assertEqual(len(prop.Columns), 3)
        col = prop.Columns[0]
        self.assertEqual(col.Name, "Temperature")
        self.assertEqual(col.Type, "Quantity")
        self.assertEqual(col.URL, "")
        self.assertEqual(col.Units, "C")
        self.assertEqual(col.Description, "Temperature")
        col = prop.Columns[1]
        self.assertEqual(col.Name, "Density")
        self.assertEqual(col.Type, "Quantity")
        self.assertEqual(col.URL, "https://en.wikipedia.org/wiki/Density")
        self.assertEqual(col.Units, "kg/m^3")
        self.assertEqual(col.Description, "Density in [FreeCAD Density unit]")
        col = prop.Columns[2]
        self.assertEqual(col.Name, "InitialYieldStress")
        self.assertEqual(col.Type, "Quantity")
        self.assertEqual(col.URL, "")
        self.assertEqual(col.Units, "kPa")
        self.assertEqual(col.Description, "Saturation stress for Voce isotropic hardening [FreeCAD Pressure unit]\n")

        prop = model.Properties["TestArray3D"]
        self.assertEqual(prop.Name, "TestArray3D")
        self.assertEqual(prop.Type, "3DArray")
        self.assertEqual(prop.URL, "")
        self.assertEqual(prop.Units, "")
        self.assertEqual(prop.Description, "3 Dimensional array showing stress and strain as a function of temperature\n")
        self.assertEqual(len(prop.Columns), 3)
        col = prop.Columns[0]
        self.assertEqual(col.Name, "Temperature")
        self.assertEqual(col.Type, "Quantity")
        self.assertEqual(col.URL, "")
        self.assertEqual(col.Units, "C")
        self.assertEqual(col.Description, "Temperature")
        col = prop.Columns[1]
        self.assertEqual(col.Name, "Stress")
        self.assertEqual(col.Type, "Quantity")
        self.assertEqual(col.URL, "")
        self.assertEqual(col.Units, "MPa")
        self.assertEqual(col.Description, "Stress")
        col = prop.Columns[2]
        self.assertEqual(col.Name, "Strain")
        self.assertEqual(col.Type, "Quantity")
        self.assertEqual(col.URL, "")
        self.assertEqual(col.Units, "MPa")
        self.assertEqual(col.Description, "Strain")

    def testModelInheritance(self):
        """ Test that the inherited models have been loaded correctly """
        model = self.ModelManager.getModel(self.uuids.LinearElastic)
        self.assertIsNotNone(model)
        self.assertEqual(model.Name, "Linear Elastic")
        self.assertEqual(model.UUID, "7b561d1d-fb9b-44f6-9da9-56a4f74d7536")
        self.assertIn("Density", model.Properties)
        prop = model.Properties["Density"]
        self.assertEqual(prop.Name, "Density")
        self.assertEqual(prop.Type, "Quantity")
        self.assertEqual(prop.URL, "https://en.wikipedia.org/wiki/Density")
        self.assertEqual(prop.Units, "kg/m^3")
        self.assertEqual(prop.Description, "Density in [FreeCAD Density unit]")
        prop = model.Properties["BulkModulus"]
        self.assertEqual(prop.Name, "BulkModulus")
        self.assertEqual(prop.DisplayName, "Bulk Modulus")
        self.assertEqual(prop.Type, "Quantity")
        self.assertEqual(prop.URL, "https://en.wikipedia.org/wiki/Bulk_modulus")
        self.assertEqual(prop.Units, "kPa")
        self.assertEqual(prop.Description, "Bulk modulus in [FreeCAD Pressure unit]")
        prop = model.Properties["PoissonRatio"]
        self.assertEqual(prop.Name, "PoissonRatio")
        self.assertEqual(prop.DisplayName, "Poisson Ratio")
        self.assertEqual(prop.Type, "Float")
        self.assertEqual(prop.URL, "https://en.wikipedia.org/wiki/Poisson%27s_ratio")
        self.assertEqual(prop.Units, "")
        self.assertEqual(prop.Description, "Poisson's ratio [unitless]")
        prop = model.Properties["ShearModulus"]
        self.assertEqual(prop.Name, "ShearModulus")
        self.assertEqual(prop.DisplayName, "Shear Modulus")
        self.assertEqual(prop.Type, "Quantity")
        self.assertEqual(prop.URL, "https://en.wikipedia.org/wiki/Shear_modulus")
        self.assertEqual(prop.Units, "kPa")
        self.assertEqual(prop.Description, "Shear modulus in [FreeCAD Pressure unit]")
        prop = model.Properties["YoungsModulus"]
        self.assertEqual(prop.Name, "YoungsModulus")
        self.assertEqual(prop.DisplayName, "Young's Modulus")
        self.assertEqual(prop.Type, "Quantity")
        self.assertEqual(prop.URL, "https://en.wikipedia.org/wiki/Young%27s_modulus")
        self.assertEqual(prop.Units, "kPa")
        self.assertEqual(prop.Description, "Young's modulus (or E-Module) in [FreeCAD Pressure unit]")


# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2022 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

"""Defines the Addon class to encapsulate information about FreeCAD Addons"""

import datetime
import importlib.util
import os
import re
from urllib.parse import urlparse, urlunparse
from typing import Dict, Set, List, Optional
from threading import Lock
from enum import IntEnum, auto
import xml.etree.ElementTree

try:
    import importlib.metadata as importlib_metadata
except ImportError:
    try:
        import importlib_metadata
    except ImportError:
        importlib_metadata = None

import addonmanager_freecad_interface as fci
from addonmanager_macro import Macro
import addonmanager_utilities as utils
from addonmanager_metadata import (
    Metadata,
    MetadataReader,
    UrlType,
    Version,
    DependencyType,
)
from AddonStats import AddonStats

translate = fci.translate

#  A list of internal workbenches that can be used as a dependency of an Addon
INTERNAL_WORKBENCHES = {
    "assembly": "Assembly",
    "bim": "BIM",
    "cam": "CAM",
    "draft": "Draft",
    "fem": "FEM",
    "import": "Import",
    "material": "Material",
    "mesh": "Mesh",
    "meshpart": "MeshPart",
    "openscad": "OpenSCAD",
    "part": "Part",
    "partdesign": "PartDesign",
    "plot": "Plot",
    "points": "Points",
    "reverseengineering": "ReverseEngineering",
    "robot": "Robot",
    "sketcher": "Sketcher",
    "spreadsheet": "Spreadsheet",
    "techdraw": "TechDraw",
    "tux": "Tux",
    "web": "Web",
}


class Addon:
    """Encapsulates information about a FreeCAD addon"""

    class Kind(IntEnum):
        """The type of Addon: Workbench, macro, or package"""

        WORKBENCH = 1
        MACRO = 2
        PACKAGE = 3

        def __str__(self) -> str:
            if self.value == 1:
                return "Workbench"
            if self.value == 2:
                return "Macro"
            if self.value == 3:
                return "Package"
            return "ERROR_TYPE"

    class Status(IntEnum):
        """The installation status of an Addon"""

        NOT_INSTALLED = 0
        UNCHECKED = 1
        NO_UPDATE_AVAILABLE = 2
        UPDATE_AVAILABLE = 3
        PENDING_RESTART = 4
        CANNOT_CHECK = 5  # Probably now obsolete  TODO: remove?
        UNKNOWN = 100

        def __lt__(self, other):
            if self.__class__ is other.__class__:
                return self.value < other.value
            return NotImplemented

        def __le__(self, other):
            return self < other or self == other

        def __str__(self) -> str:
            if self.value == 0:
                result = "Not installed"
            elif self.value == 1:
                result = "Unchecked"
            elif self.value == 2:
                result = "No update available"
            elif self.value == 3:
                result = "Update available"
            elif self.value == 4:
                result = "Restart required"
            elif self.value == 5:
                result = "Can't check"
            else:
                result = "ERROR_STATUS"
            return result

    class Dependencies:
        """Addon dependency information"""

        def __init__(self):
            self.required_external_addons: List["Addon"] = []
            self.blockers: List["Addon"] = []
            self.replaces: List["Addon"] = []
            self.internal_workbenches: Set[str] = set()  # Required internal workbenches
            self.python_requires: Set[str] = set()
            self.python_optional: Set[str] = set()
            self.python_min_version = Version(from_list=[3, 0, 0])

    class DependencyType(IntEnum):
        """What kind of dependency a given dependency is"""

        INTERNAL_WORKBENCH = auto()
        REQUIRED_ADDON = auto()
        BLOCKED_ADDON = auto()
        REPLACED_ADDON = auto()
        REQUIRED_PYTHON = auto()
        OPTIONAL_PYTHON = auto()

    class ResolutionFailed(RuntimeError):
        """An exception type for dependency resolution failure."""

    # The location of Addon Manager cache files: overridden by testing code
    cache_directory = os.path.join(fci.DataPaths().cache_dir, "AddonManager")

    # The location of the Mod directory: overridden by testing code
    mod_directory = fci.DataPaths().mod_dir

    # The location of the Macro directory: overridden by testing code
    macro_directory = fci.DataPaths().macro_dir

    def __init__(
        self,
        name: str,
        url: str = "",
        status: Status = Status.UNKNOWN,
        branch: str = "",
    ):
        self.name = name.strip()
        self.display_name = self.name
        self.url = url.strip()
        self.relative_cache_path = ""
        self.branch = branch.strip()
        self.branch_display_name = branch.strip()
        self.repo_type = Addon.Kind.WORKBENCH
        self.description = None
        self.tags = set()  # Just a cache, loaded from Metadata
        self.remote_last_updated: Optional[datetime.datetime] = None
        self.stats = AddonStats()
        self.curated = True
        self.score = 0

        # In cases where there are multiple versions/branches/installations available for an addon,
        # this dictionary is the mapping from the displayed name in the UI (as given in the
        # catalog) to the Addon object.
        self.sub_addons = {}

        # To prevent multiple threads from running git actions on this repo at the
        # same time
        self.git_lock = Lock()

        # To prevent multiple threads from accessing the status at the same time
        self.status_lock = Lock()
        self.update_status = status

        self._clean_url()

        self.metadata: Optional[Metadata] = None
        self.icon = None  # A QIcon version of this Addon's icon
        self.icon_data: bytes = bytes()  # In-memory version of this icon's data
        self.macro = None  # Bridge to Gaël Écorchard's macro management class
        self.updated_timestamp = None
        self.installed_version = None
        self.installed_metadata = None

        # Each repo is also a node in a directed dependency graph (referenced by name so
        # they can be serialized):
        self.requires: Set[str] = set()
        self.blocks: Set[str] = set()

        # And maintains a list of required and optional Python dependencies
        self.python_requires: Set[str] = set()
        self.python_optional: Set[str] = set()
        self.python_min_version = Version(from_list=[3, 0, 0])

        self._icon_file = None
        self._cached_license: str = ""
        self._cached_update_date = None

    def __eq__(self, other):
        if not isinstance(other, Addon):
            return NotImplemented
        return self.name == other.name and self.branch_display_name == other.branch_display_name

    def __lt__(self, other):
        if not isinstance(other, Addon):
            return NotImplemented
        if self.name == other.name:
            return self.branch_display_name < other.branch_display_name
        return self.name < other.name

    def __le__(self, other):
        if not isinstance(other, Addon):
            return NotImplemented
        return self == other or self < other

    def __hash__(self):
        return hash((self.name, self.branch_display_name))

    def _clean_url(self):
        # The url should never end in ".git", so strip it if it's there
        parsed_url = urlparse(self.url)
        if parsed_url.path.endswith(".git"):
            self.url = parsed_url.scheme + "://" + parsed_url.netloc + parsed_url.path[:-4]
            if parsed_url.query:
                self.url += "?" + parsed_url.query
            if parsed_url.fragment:
                self.url += "#" + parsed_url.fragment

    def __str__(self) -> str:
        result = f"FreeCAD {self.repo_type}\n"
        result += f"Name: {self.name}\n"
        result += f"URL: {self.url}\n"
        result += "Has metadata\n" if self.metadata is not None else "No metadata found\n"
        if self.macro is not None:
            result += "Has linked Macro object\n"
        return result

    @property
    def license(self):
        if not self._cached_license:
            self._cached_license = "UNLICENSED"
            if self.metadata and self.metadata.license:
                self._cached_license = self.metadata.license
            elif self.stats and self.stats.license:
                self._cached_license = self.stats.license
            elif self.macro:
                if self.macro.license:
                    self._cached_license = self.macro.license
                elif self.macro.on_wiki:
                    self._cached_license = "CC-BY-3.0"
        return self._cached_license

    @property
    def update_date(self):
        if self.remote_last_updated is not None:
            return self.remote_last_updated
        if self._cached_update_date is None:
            self._cached_update_date = 0
            if self.stats and self.stats.last_update_time:
                self._cached_update_date = self.stats.last_update_time
            elif self.macro and self.macro.date:
                # Try to parse the date:
                try:
                    self._cached_update_date = utils.process_date_string_to_python_datetime(
                        self.macro.date
                    )
                except ValueError as e:
                    fci.Console.PrintWarning(str(e) + "\n")
            else:
                fci.Console.PrintWarning(f"No update date info for {self.name}\n")
        return self._cached_update_date

    @classmethod
    def from_macro(cls, macro: Macro):
        """Create an Addon object from a Macro wrapper object"""

        if macro.is_installed():
            status = Addon.Status.UNCHECKED
        else:
            status = Addon.Status.NOT_INSTALLED
        instance = Addon(macro.name, macro.url, status, "master")
        instance.macro = macro
        instance.repo_type = Addon.Kind.MACRO
        instance.description = macro.desc
        return instance

    def load_metadata_file(self, file: str) -> None:
        """Read a given metadata file and set it as this object's metadata"""

        if os.path.exists(file):
            try:
                metadata = MetadataReader.from_file(file)
            except xml.etree.ElementTree.ParseError:
                fci.Console.PrintWarning(
                    "An invalid or corrupted package.xml file was found in the cache for"
                )
                fci.Console.PrintWarning(f" {self.name}... ignoring the bad data.\n")
                return
            self.set_metadata(metadata)
            self._clean_url()
        else:
            fci.Console.PrintLog(f"Internal error: {file} does not exist")

    def _load_installed_metadata(self) -> None:
        # If it is actually installed, there is a SECOND metadata file, in the actual installation,
        # that may not match the cached one if the Addon has not been updated but the cache has.
        mod_dir = os.path.join(self.mod_directory, self.name)
        installed_metadata_path = os.path.join(mod_dir, "package.xml")
        if os.path.isfile(installed_metadata_path):
            try:
                self.installed_metadata = MetadataReader.from_file(installed_metadata_path)
            except xml.etree.ElementTree.ParseError:
                fci.Console.PrintWarning(
                    "An invalid or corrupted package.xml file was found in installation of"
                )
                fci.Console.PrintWarning(f" {self.name}... ignoring the bad data.\n")
                return

    def set_metadata(self, metadata: Metadata) -> None:
        """Set the given metadata object as this object's metadata, updating the
        object's display name and package type information to match, as well as
        updating any dependency information, etc.
        """

        self.metadata = metadata
        self.display_name = (
            str(metadata.name)
            .replace("\n", " ")
            .replace("\r", " ")
            .replace("{", "")
            .replace("}", "")
        )
        self.repo_type = Addon.Kind.PACKAGE
        self.description = metadata.description
        for url in metadata.url:
            if url.type == UrlType.repository:
                self.url = url.location
                self.branch = url.branch if url.branch else "master"
        self._clean_url()
        self.extract_tags(self.metadata)
        self.extract_metadata_dependencies(self.metadata)

    @staticmethod
    def version_is_ok(metadata: Metadata) -> bool:
        """Checks to see if the current running version of FreeCAD meets the
        requirements set by the passed-in metadata parameter."""

        from_fci = list(fci.Version())
        fc_version = Version(from_list=from_fci)

        dep_fc_min = metadata.freecadmin if metadata.freecadmin else fc_version
        dep_fc_max = metadata.freecadmax if metadata.freecadmax else fc_version

        return dep_fc_min <= fc_version <= dep_fc_max

    def extract_metadata_dependencies(self, metadata: Metadata):
        """Read dependency information from a metadata object and store it in this
        Addon"""

        # Version check: if this piece of metadata doesn't apply to this version of
        # FreeCAD, just skip it.
        if not Addon.version_is_ok(metadata):
            return

        if metadata.pythonmin:
            self.python_min_version = Version(from_list=metadata.pythonmin.version_as_list)

        for dep in metadata.depend:
            if dep.dependency_type == DependencyType.internal:
                if dep.package.strip().lower() in INTERNAL_WORKBENCHES:
                    self.requires.add(dep.package)
                else:
                    fci.Console.PrintWarning(
                        translate(
                            "AddonsInstaller",
                            "{}: Unrecognized internal workbench '{}'\n",
                        ).format(self.name, dep.package)
                    )
            elif dep.dependency_type == DependencyType.addon:
                self.requires.add(dep.package)
            elif dep.dependency_type == DependencyType.python:
                if dep.optional:
                    self.python_optional.add(dep.package)
                else:
                    self.python_requires.add(dep.package)
            else:
                # Automatic resolution happens later, once we have a complete list of
                # Addons
                self.requires.add(dep.package)

        for dep in metadata.conflict:
            self.blocks.add(dep.package)

        # Recurse
        content = metadata.content
        for _, value in content.items():
            for item in value:
                self.extract_metadata_dependencies(item)

    def verify_url_and_branch(self, url: str, branch: str) -> None:
        """Print diagnostic information for Addon Developers if their metadata is
        inconsistent with the actual fetch location. Most often this is due to using
        the wrong branch name."""

        if self.url != url:
            fci.Console.PrintWarning(
                translate(
                    "AddonsInstaller",
                    "Addon Developer Warning: Repository URL set in package.xml file for addon {} ({}) does not match the URL it was fetched from ({})",
                ).format(self.display_name, self.url, url)
                + "\n"
            )
        if self.branch != branch:
            fci.Console.PrintWarning(
                translate(
                    "AddonsInstaller",
                    "Addon Developer Warning: Repository branch set in package.xml file for addon {} ({}) does not match the branch it was fetched from ({})",
                ).format(self.display_name, self.branch, branch)
                + "\n"
            )

    def extract_tags(self, metadata: Metadata) -> None:
        """Read the tags from the metadata object"""

        # Version check: if this piece of metadata doesn't apply to this version of
        # FreeCAD, just skip it.
        if not Addon.version_is_ok(metadata):
            return

        for new_tag in metadata.tag:
            self.tags.add(new_tag)

        content = metadata.content
        for _, value in content.items():
            for item in value:
                self.extract_tags(item)

    def contains_workbench(self) -> bool:
        """Determine if this package contains (or is) a workbench"""

        if self.repo_type == Addon.Kind.WORKBENCH:
            return True
        return self.contains_packaged_content("workbench")

    def contains_macro(self) -> bool:
        """Determine if this package contains (or is) a macro"""

        if self.repo_type == Addon.Kind.MACRO:
            return True
        return self.contains_packaged_content("macro")

    def contains_packaged_content(self, content_type: str):
        """Determine if the package contains content_type"""
        if self.repo_type == Addon.Kind.PACKAGE:
            if self.metadata is None:
                fci.Console.PrintLog(
                    f"Addon Manager internal error: lost metadata for package {self.name}\n"
                )
                return False
            content = self.metadata.content
            return content_type in content
        return False

    def contains_preference_pack(self) -> bool:
        """Determine if this package contains a preference pack"""
        return self.contains_packaged_content("preferencepack")

    def contains_bundle(self) -> bool:
        """Determine if this package contains a bundle"""
        return self.contains_packaged_content("bundle")

    def contains_other(self) -> bool:
        """Determine if this package contains an "other" content item"""
        return self.contains_packaged_content("other")

    def walk_dependency_tree(self, all_repos: Dict[str, "Addon"], deps: Dependencies):
        """Compute the total dependency tree for this repo (recursive)
        - all_repos is a dictionary of repos, keyed on the name of the repo
        - deps is an Addon.Dependency object encapsulating all the types of dependency
        information that may be needed.
        """

        deps.python_requires |= self.python_requires
        deps.python_optional |= self.python_optional
        deps.python_min_version = max(self.python_min_version, deps.python_min_version)

        for dep in self.requires:
            if dep in all_repos:
                if dep not in deps.required_external_addons:
                    deps.required_external_addons.append(all_repos[dep])
                    all_repos[dep].walk_dependency_tree(all_repos, deps)
            else:
                # See if this is an internal workbench:
                if dep.upper().endswith("WB"):
                    real_name = dep[:-2].strip().lower()
                elif dep.upper().endswith("WORKBENCH"):
                    real_name = dep[:-9].strip().lower()
                else:
                    real_name = dep.strip().lower()

                if real_name in INTERNAL_WORKBENCHES:
                    deps.internal_workbenches.add(INTERNAL_WORKBENCHES[real_name])
                else:
                    # Assume it's a Python requirement of some kind:
                    deps.python_requires.add(dep)

        for dep in self.blocks:
            if dep in all_repos:
                if all_repos[dep] not in deps.blockers:
                    deps.blockers.append(all_repos[dep])

    def status(self):
        """Threadsafe access to the current update status"""
        with self.status_lock:
            return self.update_status

    def set_status(self, status):
        """Threadsafe setting of the update status"""
        with self.status_lock:
            self.update_status = status

    def is_disabled(self):
        """Check to see if the disabling stopfile exists"""

        stopfile = os.path.join(self.mod_directory, self.name, "ADDON_DISABLED")
        return os.path.exists(stopfile)

    def disable(self):
        """Disable this addon from loading when FreeCAD starts up by creating a
        stopfile"""

        stopfile = os.path.join(self.mod_directory, self.name, "ADDON_DISABLED")
        with open(stopfile, "w", encoding="utf-8") as f:
            f.write(
                "The existence of this file prevents FreeCAD from loading this Addon. To re-enable, delete the file."
            )

        if self.contains_workbench():
            self.disable_workbench()

    def enable(self):
        """Re-enable loading this addon by deleting the stopfile"""

        stopfile = os.path.join(self.mod_directory, self.name, "ADDON_DISABLED")
        try:
            os.unlink(stopfile)
        except FileNotFoundError:
            # If the file disappeared on us, there's no need to do anything as it's gone already
            pass

        if self.contains_workbench():
            self.enable_workbench()

    def enable_workbench(self):
        workbench_name = self.get_workbench_name()

        # Remove from the list of disabled.
        self.remove_from_disabled_wbs(workbench_name)

    def disable_workbench(self):
        pref = fci.ParamGet("User parameter:BaseApp/Preferences/Workbenches")
        workbench_name = self.get_workbench_name()

        # Add the wb to the list of disabled if it was not already
        disabled_wbs = pref.GetString(
            "Disabled",
            "NoneWorkbench,TestWorkbench,InspectionWorkbench,RobotWorkbench,OpenSCADWorkbench",
        )
        # print(f"start disabling {disabled_wbs}")
        disabled_wbs_list = disabled_wbs.split(",")
        if not (workbench_name in disabled_wbs_list):
            disabled_wbs += "," + workbench_name
        pref.SetString("Disabled", disabled_wbs)
        # print(f"done disabling :  {disabled_wbs} \n")

    def remove_workbench(self):
        pref = fci.ParamGet("User parameter:BaseApp/Preferences/Workbenches")
        workbench_name = self.get_workbench_name()

        # Remove from the list of ordered.
        ordered_wbs = pref.GetString("Ordered", "")
        ordered_wbs_list = ordered_wbs.split(",")
        ordered_wbs = ""
        for wb in ordered_wbs_list:
            if wb != workbench_name:
                if ordered_wbs != "":
                    ordered_wbs += ","
                ordered_wbs += wb
        pref.SetString("Ordered", ordered_wbs)

        # Remove from the list of disabled.
        self.remove_from_disabled_wbs(workbench_name)

    @staticmethod
    def remove_from_disabled_wbs(workbench_name: str):
        pref = fci.ParamGet("User parameter:BaseApp/Preferences/Workbenches")

        disabled_wbs = pref.GetString("Disabled", "NoneWorkbench,TestWorkbench")
        disabled_wbs_list = disabled_wbs.split(",")
        disabled_wbs = ""
        for wb in disabled_wbs_list:
            if wb != workbench_name:
                if disabled_wbs != "":
                    disabled_wbs += ","
                disabled_wbs += wb
        pref.SetString("Disabled", disabled_wbs)

    def get_workbench_name(self) -> str:
        """Find the name of the workbench class (i.e., the name under which it's
        registered in FreeCAD core)"""
        wb_name = ""

        if self.repo_type == Addon.Kind.PACKAGE:
            for wb in self.metadata.content["workbench"]:  # we may have more than one wb.
                if wb_name != "":
                    wb_name += ","
                wb_name += wb.classname
        if self.repo_type == Addon.Kind.WORKBENCH or wb_name == "":
            wb_name = self.try_find_workbench_name_in_files()
        if wb_name == "":
            wb_name = self.name
        return wb_name

    def try_find_workbench_name_in_files(self) -> str:
        """Attempt to locate a line with an addWorkbench command in the workbench's
        Python files. If it is directly instantiating a workbench, then we can use
        the line to determine the classname for this workbench. If it uses a variable,
        or if the line doesn't exist at all, an empty string is returned."""
        mod_dir = os.path.join(self.mod_directory, self.name)

        for root, _, files in os.walk(mod_dir):
            for f in files:
                current_file = os.path.join(root, f)
                if not os.path.isdir(current_file):
                    filename, extension = os.path.splitext(current_file)
                    if extension == ".py":
                        wb_classname = self._find_classname_in_file(current_file)
                        if wb_classname:
                            return wb_classname
        return ""

    @staticmethod
    def _find_classname_in_file(current_file) -> str:
        try:
            with open(current_file, "r", encoding="utf-8") as python_file:
                content = python_file.read()
                search_result = re.search(r"Gui.addWorkbench\s*\(\s*(\w+)\s*\(\s*\)\s*\)", content)
                if search_result:
                    return search_result.group(1)
        except OSError:
            # Fall through to the classname-not-found case (if we couldn't read the file, etc.)
            pass
        return ""

    def get_zip_url(self) -> str:
        if self.url.endswith(".zip"):
            zip_url = self.url
        else:
            # The ZIP url is based on the location of the main cache file:
            if self.relative_cache_path:
                cache_file_url = fci.Preferences().get("addon_catalog_cache_url")
                parsed_url = urlparse(cache_file_url)
                path_parts = parsed_url.path.rpartition("/")
                new_path = path_parts[0] + "/" + self.relative_cache_path
                zip_url = urlunparse(
                    (
                        parsed_url.scheme,
                        parsed_url.netloc,
                        new_path,
                        parsed_url.params,
                        parsed_url.query,
                        parsed_url.fragment,
                    )
                )
            else:
                zip_url = utils.get_zip_url(self)
        return zip_url


# @dataclass(frozen)
class MissingDependencies:
    """Encapsulates a group of four types of dependencies:
    * Internal workbenches -> wbs
    * External addons -> external_addons
    * Required Python packages -> python_requires
    * Optional Python packages -> python_optional
    """

    def __init__(self):
        self.external_addons: List[Addon] = []
        self.wbs: List[str] = []
        self.python_requires: List[str] = []
        self.python_optional: List[str] = []
        self.python_min_version = Version(from_list=[3, 0, 0])

    def import_from_addon(self, repo: Addon, all_repos: List[Addon]):
        deps = Addon.Dependencies()
        repo_name_dict = {}
        for r in all_repos:
            repo_name_dict[r.name] = r
            if hasattr(r, "display_name"):
                # Test harness might not provide a display name
                repo_name_dict[r.display_name] = r

        if hasattr(repo, "walk_dependency_tree"):
            # Sometimes the test harness doesn't provide this function to override
            # any dependency checking
            repo.walk_dependency_tree(repo_name_dict, deps)

        for dep in deps.required_external_addons:
            if dep.status() == Addon.Status.NOT_INSTALLED:
                self.external_addons.append(dep)

        # Now check the loaded addons to see if we are missing an internal workbench:
        if fci.FreeCADGui:
            wbs = [wb.lower() for wb in fci.FreeCADGui.listWorkbenches()]
        else:
            wbs = []

        for dep in deps.internal_workbenches:
            if dep.lower() + "workbench" not in wbs:
                if dep.lower() == "plot":
                    # Special case for plot, which is no longer a full workbench:
                    try:
                        __import__("Plot")
                    except ImportError:
                        # Plot might fail for a number of reasons
                        self.wbs.append(dep)
                        fci.Console.PrintLog("Failed to import Plot module\n")
                elif dep.lower() == "meshpart":
                    # MeshPart is strange: it doesn't ever appear in the listWorkbenches() output
                    try:
                        __import__("MeshPart")
                    except ImportError:
                        self.wbs.append(dep)
                        fci.Console.PrintLog("Failed to import MeshPart module\n")
                else:
                    self.wbs.append(dep)

        # Check the Python dependencies:

        # Python version:
        self.python_min_version = max(self.python_min_version, deps.python_min_version)

        # Required packages -- only add if it's not in the list already and is not installed
        for py_dep in deps.python_requires:
            if py_dep not in self.python_requires and not self.package_is_installed(py_dep):
                self.python_requires.append(py_dep)

        # Optional packages -- only add if it's not in the list already and is not installed
        for py_dep in deps.python_optional:
            if py_dep not in self.python_optional and not self.package_is_installed(py_dep):
                self.python_optional.append(py_dep)

        self.wbs.sort()
        self.external_addons.sort()
        self.python_requires.sort()
        self.python_optional.sort()

        # Something on the optional list *and* the required list should be removed from
        # optional (since it's *not* optional)
        self.python_optional = [
            option for option in self.python_optional if option not in self.python_requires
        ]

    def join(self, other: "MissingDependencies"):
        """Join two sets of missing dependencies together"""
        self.external_addons.extend(
            [x for x in other.external_addons if x not in self.external_addons]
        )
        self.wbs.extend([x for x in other.wbs if x not in self.wbs])
        self.python_requires.extend(
            [x for x in other.python_requires if x not in self.python_requires]
        )
        self.python_optional.extend(
            [x for x in other.python_optional if x not in self.python_optional]
        )
        self.python_min_version = max(self.python_min_version, other.python_min_version)

        # Clean up optional:
        self.python_optional = [x for x in self.python_optional if x not in self.python_requires]

    @staticmethod
    def package_is_installed(package_name: str) -> bool:
        """Check to see if a Python package is installed (i.e., if it can be imported).

        Returns False if the running version of Python can't check the metadata for the package, and
        the package can't be directly imported by its given name (e.g. `python-distutils`, which
        gets imported as `distutils` instead, so won't match the simple test). This only applies to
        Python < 3.8.

        :param package_name: The PyPI name of the package to check for.
        :return: True if the package is installed, False otherwise."""

        # The simplest test: can we import it with the stated dependency name?
        if importlib.util.find_spec(package_name) is not None:
            return True

        # On Python 3.8 and later, or if the importlib_metadata package is installed, we
        # can do the check by PyPI package name:
        if importlib_metadata is None:
            fci.Console.PrintMessage(
                f"Cannot check for installation of `{package_name}`... marking it for "
                "reinstallation to be safe\n"
            )
            return False

        try:
            # Only the side effect matters here: if this call succeeds, the package is installed.
            # If an exception is raised, it is not.
            _ = importlib_metadata.distribution(package_name)
        except importlib_metadata.PackageNotFoundError:
            return False

        return True


def cycle_to_sub_addon(original: Addon, sub_addon: Addon, addon_model):
    """Given an addon with sub-addons, cycle the sub-addon to be the primary. After this call,
    the addon_list will contain the sub_addon, which will itself have a list of sub-addons that
    now includes the original addon (and *not* itself).
    :param original: The addon that is currently the primary
    :param sub_addon: The sub-addon that should be the primary
    :param addon_model: The PackageListItemModel containing all addons"""

    addon_model.remove_item(original)

    new_sub_dict = {original.branch_display_name: original}
    for key, value in original.sub_addons.items():
        if key != sub_addon.branch_display_name:
            new_sub_dict[key] = value

    sub_addon.sub_addons = new_sub_dict
    original.sub_addons = {}

    addon_model.append_item(sub_addon)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides various functions to work with fillets."""
## @package fillets
# \ingroup draftgeoutils
# \brief Provides various functions to work with fillets.

import math
import lazy_loader.lazy_loader as lz

import FreeCAD as App

from draftgeoutils.general import precision
from draftgeoutils.arcs import arcFrom2Pts
from draftgeoutils.wires import isReallyClosed

# Delay import of module until first use because it is heavy
Part = lz.LazyLoader("Part", globals(), "Part")

## \addtogroup draftgeoutils
# @{


def fillet(lEdges, r, chamfer=False):
    """Return a list of sorted edges describing a round corner.

    Author: Jacques-Antoine Gaudin
    """

    def getCurveType(edge, existingCurveType=None):
        """Build or complete a dictionary containing edges.

        The dictionary contains edges with keys 'Arc' and 'Line'.
        """
        if not existingCurveType:
            existingCurveType = {"Line": [], "Arc": []}
        if issubclass(type(edge.Curve), Part.LineSegment):
            existingCurveType["Line"] += [edge]
        elif issubclass(type(edge.Curve), Part.Line):
            existingCurveType["Line"] += [edge]
        elif issubclass(type(edge.Curve), Part.Circle):
            existingCurveType["Arc"] += [edge]
        else:
            raise ValueError("Edge's curve must be either Line or Arc")
        return existingCurveType

    rndEdges = lEdges[0:2]
    rndEdges = Part.__sortEdges__(rndEdges)

    if len(rndEdges) < 2:
        return rndEdges

    if r <= 0:
        print("DraftGeomUtils.fillet: Error: radius is negative.")
        return rndEdges

    curveType = getCurveType(rndEdges[0])
    curveType = getCurveType(rndEdges[1], curveType)

    # Part.__sortEdges__() does not reverse edges. There is no guarantee that
    # the endpoint of the 1st edge is the corner point.
    edge1_sta, edge1_end = [rndEdges[0].Vertexes[i].Point for i in [0, -1]]
    edge2_sta, edge2_end = [rndEdges[1].Vertexes[i].Point for i in [0, -1]]
    tol = 1e-7
    if edge1_sta.isEqual(edge2_sta, tol):
        lVertexes = [rndEdges[0].Vertexes[-1], rndEdges[0].Vertexes[0], rndEdges[1].Vertexes[-1]]
    elif edge1_sta.isEqual(edge2_end, tol):
        lVertexes = [rndEdges[0].Vertexes[-1], rndEdges[0].Vertexes[0], rndEdges[1].Vertexes[0]]
    elif edge1_end.isEqual(edge2_sta, tol):
        lVertexes = [rndEdges[0].Vertexes[0], rndEdges[0].Vertexes[-1], rndEdges[1].Vertexes[-1]]
    else:
        lVertexes = [rndEdges[0].Vertexes[0], rndEdges[0].Vertexes[-1], rndEdges[1].Vertexes[0]]

    if len(curveType["Line"]) == 2:
        # Deals with 2-line-edges lists
        U1 = lVertexes[0].Point.sub(lVertexes[1].Point)
        U1.normalize()

        U2 = lVertexes[2].Point.sub(lVertexes[1].Point)
        U2.normalize()

        alpha = U1.getAngle(U2)

        # Edges have same direction
        if round(alpha, precision()) == 0 or round(alpha - math.pi, precision()) == 0:
            print("DraftGeomUtils.fillet: Warning: " "edges have same direction. Did nothing")
            return rndEdges

        dToCenter = r / math.sin(alpha / 2.0)
        dToTangent = (dToCenter**2 - r**2) ** (0.5)
        dirVect = App.Vector(U1)
        dirVect.scale(dToTangent, dToTangent, dToTangent)
        arcPt1 = lVertexes[1].Point.add(dirVect)

        dirVect = U2.add(U1)
        dirVect.normalize()
        dirVect.scale(dToCenter - r, dToCenter - r, dToCenter - r)
        arcPt2 = lVertexes[1].Point.add(dirVect)

        dirVect = App.Vector(U2)
        dirVect.scale(dToTangent, dToTangent, dToTangent)
        arcPt3 = lVertexes[1].Point.add(dirVect)

        if (dToTangent > rndEdges[0].Length) or (dToTangent > rndEdges[1].Length):
            print("DraftGeomUtils.fillet: Error: radius value ", r, " is too high")
            return rndEdges

        if chamfer:
            rndEdges[1] = Part.Edge(Part.LineSegment(arcPt1, arcPt3))
        else:
            rndEdges[1] = Part.Edge(Part.Arc(arcPt1, arcPt2, arcPt3))

        if lVertexes[0].Point == arcPt1:
            # fillet consumes entire first edge
            rndEdges.pop(0)
        else:
            rndEdges[0] = Part.Edge(Part.LineSegment(lVertexes[0].Point, arcPt1))

        if lVertexes[2].Point != arcPt3:
            # fillet does not consume entire second edge
            rndEdges += [Part.Edge(Part.LineSegment(arcPt3, lVertexes[2].Point))]

        return rndEdges

    elif len(curveType["Arc"]) == 1:
        # Deals with lists containing an arc and a line
        if rndEdges[0] in curveType["Arc"]:
            lineEnd = lVertexes[2]
            arcEnd = lVertexes[0]
            arcFirst = True
        else:
            lineEnd = lVertexes[0]
            arcEnd = lVertexes[2]
            arcFirst = False
        arcCenter = curveType["Arc"][0].Curve.Center
        arcRadius = curveType["Arc"][0].Curve.Radius
        arcAxis = curveType["Arc"][0].Curve.Axis
        arcLength = curveType["Arc"][0].Length

        U1 = lineEnd.Point.sub(lVertexes[1].Point)
        U1.normalize()
        toCenter = arcCenter.sub(lVertexes[1].Point)
        if arcFirst:  # make sure the tangent points towards the arc
            T = arcAxis.cross(toCenter)
        else:
            T = toCenter.cross(arcAxis)

        projCenter = toCenter.dot(U1)
        if round(abs(projCenter), precision()) > 0:
            normToLine = U1.cross(T).cross(U1)
        else:
            normToLine = App.Vector(toCenter)
        normToLine.normalize()

        dCenterToLine = toCenter.dot(normToLine) - r

        if round(projCenter, precision()) > 0:
            newRadius = arcRadius - r
        elif round(projCenter, precision()) < 0 or (
            round(projCenter, precision()) == 0 and U1.dot(T) > 0
        ):
            newRadius = arcRadius + r
        else:
            print("DraftGeomUtils.fillet: Warning: " "edges are already tangent. Did nothing")
            return rndEdges

        toNewCent = newRadius**2 - dCenterToLine**2
        if toNewCent > 0:
            toNewCent = abs(abs(projCenter) - toNewCent ** (0.5))
        else:
            print("DraftGeomUtils.fillet: Error: radius value ", r, " is too high")
            return rndEdges

        U1.scale(toNewCent, toNewCent, toNewCent)
        normToLine.scale(r, r, r)
        newCent = lVertexes[1].Point.add(U1).add(normToLine)

        arcPt1 = lVertexes[1].Point.add(U1)
        arcPt2 = lVertexes[1].Point.sub(newCent)
        arcPt2.normalize()
        arcPt2.scale(r, r, r)
        arcPt2 = arcPt2.add(newCent)

        if newRadius == arcRadius - r:
            arcPt3 = newCent.sub(arcCenter)
        else:
            arcPt3 = arcCenter.sub(newCent)
        arcPt3.normalize()
        arcPt3.scale(r, r, r)
        arcPt3 = arcPt3.add(newCent)
        arcPt = [arcPt1, arcPt2, arcPt3]

        # Warning: In the following I used a trick for calling
        # the right element in arcPt or V:
        # arcFirst is a boolean so - not arcFirst is -0 or -1
        # list[-1] is the last element of a list and list[0] the first
        # this way I don't have to proceed tests to know the position
        # of the arc
        myTrick = not arcFirst

        V = [arcPt3]
        V += [arcEnd.Point]

        toCenter.scale(-1, -1, -1)

        delLength = arcRadius * V[0].sub(arcCenter).getAngle(toCenter)
        if delLength > arcLength or toNewCent > curveType["Line"][0].Length:
            print("DraftGeomUtils.fillet: Error: radius value ", r, " is too high")
            return rndEdges

        arcAsEdge = arcFrom2Pts(V[-arcFirst], V[-myTrick], arcCenter, arcAxis)

        V = [lineEnd.Point, arcPt1]
        lineAsEdge = Part.Edge(Part.LineSegment(V[-arcFirst], V[myTrick]))

        rndEdges[not arcFirst] = arcAsEdge
        rndEdges[arcFirst] = lineAsEdge
        if chamfer:
            rndEdges[1:1] = [Part.Edge(Part.LineSegment(arcPt[-arcFirst], arcPt[-myTrick]))]
        else:
            rndEdges[1:1] = [Part.Edge(Part.Arc(arcPt[-arcFirst], arcPt[1], arcPt[-myTrick]))]

        return rndEdges

    elif len(curveType["Arc"]) == 2:
        # Deals with lists of 2 arc-edges
        (arcCenter, arcRadius, arcAxis, arcLength, toCenter, T, newRadius) = (
            [],
            [],
            [],
            [],
            [],
            [],
            [],
        )

        for i in range(2):
            arcCenter += [curveType["Arc"][i].Curve.Center]
            arcRadius += [curveType["Arc"][i].Curve.Radius]
            arcAxis += [curveType["Arc"][i].Curve.Axis]
            arcLength += [curveType["Arc"][i].Length]
            toCenter += [arcCenter[i].sub(lVertexes[1].Point)]

        T += [arcAxis[0].cross(toCenter[0])]
        T += [toCenter[1].cross(arcAxis[1])]
        CentToCent = toCenter[1].sub(toCenter[0])
        dCentToCent = CentToCent.Length

        sameDirection = arcAxis[0].dot(arcAxis[1]) > 0
        TcrossT = T[0].cross(T[1])

        if sameDirection:
            if round(TcrossT.dot(arcAxis[0]), precision()) > 0:
                newRadius += [arcRadius[0] + r]
                newRadius += [arcRadius[1] + r]
            elif round(TcrossT.dot(arcAxis[0]), precision()) < 0:
                newRadius += [arcRadius[0] - r]
                newRadius += [arcRadius[1] - r]
            elif T[0].dot(T[1]) > 0:
                newRadius += [arcRadius[0] + r]
                newRadius += [arcRadius[1] + r]
            else:
                print("DraftGeomUtils.fillet: Warning: " "edges are already tangent. Did nothing")
                return rndEdges

        elif not sameDirection:
            if round(TcrossT.dot(arcAxis[0]), precision()) > 0:
                newRadius += [arcRadius[0] + r]
                newRadius += [arcRadius[1] - r]
            elif round(TcrossT.dot(arcAxis[0]), precision()) < 0:
                newRadius += [arcRadius[0] - r]
                newRadius += [arcRadius[1] + r]
            elif T[0].dot(T[1]) > 0:
                if arcRadius[0] > arcRadius[1]:
                    newRadius += [arcRadius[0] - r]
                    newRadius += [arcRadius[1] + r]
                elif arcRadius[1] > arcRadius[0]:
                    newRadius += [arcRadius[0] + r]
                    newRadius += [arcRadius[1] - r]
                else:
                    print("DraftGeomUtils.fillet: Warning: " "arcs are coincident. Did nothing")
                    return rndEdges
            else:
                print("DraftGeomUtils.fillet: Warning: " "edges are already tangent. Did nothing")
                return rndEdges

        if (
            newRadius[0] + newRadius[1] < dCentToCent
            or newRadius[0] - newRadius[1] > dCentToCent
            or newRadius[1] - newRadius[0] > dCentToCent
        ):
            print("DraftGeomUtils.fillet: Error: radius value ", r, " is too high")
            return rndEdges

        x = (dCentToCent**2 + newRadius[0] ** 2 - newRadius[1] ** 2) / (2 * dCentToCent)
        y = (newRadius[0] ** 2 - x**2) ** (0.5)

        CentToCent.normalize()
        toCenter[0].normalize()
        toCenter[1].normalize()
        if abs(toCenter[0].dot(toCenter[1])) != 1:
            normVect = CentToCent.cross(CentToCent.cross(toCenter[0]))
        else:
            normVect = T[0]

        normVect.normalize()
        CentToCent.scale(x, x, x)
        normVect.scale(y, y, y)
        newCent = arcCenter[0].add(CentToCent.add(normVect))
        CentToNewCent = [newCent.sub(arcCenter[0]), newCent.sub(arcCenter[1])]

        for i in range(2):
            CentToNewCent[i].normalize()
            if newRadius[i] == arcRadius[i] + r:
                CentToNewCent[i].scale(-r, -r, -r)
            else:
                CentToNewCent[i].scale(r, r, r)

        toThirdPt = lVertexes[1].Point.sub(newCent)
        toThirdPt.normalize()
        toThirdPt.scale(r, r, r)
        arcPt1 = newCent.add(CentToNewCent[0])
        arcPt2 = newCent.add(toThirdPt)
        arcPt3 = newCent.add(CentToNewCent[1])
        arcPt = [arcPt1, arcPt2, arcPt3]

        arcAsEdge = []
        for i in range(2):
            toCenter[i].scale(-1, -1, -1)
            delLength = arcRadius[i] * arcPt[-i].sub(arcCenter[i]).getAngle(toCenter[i])
            if delLength > arcLength[i]:
                print("DraftGeomUtils.fillet: Error: radius value ", r, " is too high")
                return rndEdges
            V = [arcPt[-i], lVertexes[-i].Point]
            arcAsEdge += [arcFrom2Pts(V[i - 1], V[-i], arcCenter[i], arcAxis[i])]

        rndEdges[0] = arcAsEdge[0]
        rndEdges[1] = arcAsEdge[1]
        if chamfer:
            rndEdges[1:1] = [Part.Edge(Part.LineSegment(arcPt[0], arcPt[2]))]
        else:
            rndEdges[1:1] = [Part.Edge(Part.Arc(arcPt[0], arcPt[1], arcPt[2]))]

        return rndEdges


def filletWire(aWire, r, chamfer=False):
    """Fillet each angle of a wire with r as radius.

    If chamfer is true, a `chamfer` is made instead, and `r` is the
    size of the chamfer.
    """
    edges = aWire.Edges
    edges = Part.__sortEdges__(edges)
    filEdges = [edges[0]]

    for i in range(len(edges) - 1):
        result = fillet([filEdges[-1], edges[i + 1]], r, chamfer)
        if len(result) > 2:
            filEdges[-1:] = result[0:3]
        else:
            filEdges[-1:] = result[0:2]

    if isReallyClosed(aWire):
        result = fillet([filEdges[-1], filEdges[0]], r, chamfer)
        if len(result) > 2:
            filEdges[-1:] = result[0:2]
            filEdges[0] = result[2]

    return Part.Wire(filEdges)


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2020 Werner Mayer <wmayer[at]users.sourceforge.net>     *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "BasicShapes.CommandShapes"
__author__ = "Werner Mayer"
__url__ = "https://www.freecad.org"
__doc__ = "Basic shapes"


import FreeCAD
from FreeCAD import Qt
import FreeCADGui

import Part
from . import Shapes
from . import ViewProviderShapes
import math
import sys


class CommandTube:
    """Command for creating Tube."""

    def GetResources(self):
        return {
            "MenuText": Qt.QT_TRANSLATE_NOOP("Part_Tube", "Tube"),
            "Accel": "",
            "CmdType": "AlterDoc:Alter3DView:AlterSelection",
            "Pixmap": "Part_Tube_Parametric",
            "ToolTip": Qt.QT_TRANSLATE_NOOP("Part_Tube", "Creates a tube"),
        }

    def Activated(self):
        text = FreeCAD.Qt.translate("QObject", "Create tube")
        FreeCAD.ActiveDocument.openTransaction(text)
        tube = FreeCAD.ActiveDocument.addObject("Part::FeaturePython", "Tube")
        Shapes.TubeFeature(tube)
        vp = ViewProviderShapes.ViewProviderTube(tube.ViewObject)
        activePart = FreeCADGui.activeView().getActiveObject("part")
        if activePart:
            activePart.addObject(tube)
        FreeCAD.ActiveDocument.recompute()
        vp.startDefaultEditMode(tube.ViewObject)

    def IsActive(self):
        return not FreeCAD.ActiveDocument is None


FreeCADGui.addCommand("Part_Tube", CommandTube())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2011 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD Arch Floor"
__author__ = "Yorik van Havre"
__url__ = "https://www.freecad.org"

## @package ArchFloor
#  \ingroup ARCH
#  \brief The Floor object and tools
#
#  This module provides tools to build Floor objects.
#  Floors are used to group different Arch objects situated
#  at a same level

"""This module provides tools to build Floor objects. Floors are used to group
different Arch objects situated at a same level.

The _Floor object and this module as a whole is now obsolete. It has been
superseded by the use of the BuildingPart class, set to the "Building Storey"
IfcType.
"""

import FreeCAD
import ArchCommands
import ArchIFC
import Draft
import DraftVecUtils

from draftutils import params

if FreeCAD.GuiUp:
    from PySide.QtCore import QT_TRANSLATE_NOOP
    import FreeCADGui
    from draftutils.translate import translate
else:
    # \cond
    def translate(ctxt, txt):
        return txt

    def QT_TRANSLATE_NOOP(ctxt, txt):
        return txt

    # \endcond


def makeFloor(objectslist=None, baseobj=None, name=None):
    """Obsolete, superseded by ArchBuildingPart.makeFloor.

    Create a floor.

    Create a new floor based on a group, and then adds the objects in
    objectslist to the new floor.

    Parameters
    ----------
    objectslist: list of <App::DocumentObject>, optional
        The objects to add to the new floor.
    baseobj:
        Unused.
    name: str, optional
        The Label for the new floor.

    Returns
    -------
    <App::DocumentObjectGroupPython>
        The created floor.
    """

    if not FreeCAD.ActiveDocument:
        FreeCAD.Console.PrintError("No active document. Aborting\n")
        return
    obj = FreeCAD.ActiveDocument.addObject("App::DocumentObjectGroupPython", "Floor")
    obj.Label = name if name else translate("Arch", "Floor")
    _Floor(obj)
    if FreeCAD.GuiUp:
        _ViewProviderFloor(obj.ViewObject)
    if objectslist:
        obj.Group = objectslist
    return obj


class _CommandFloor:
    """The command definition for the Arch workbench's gui tool, Arch Floor.

    A tool for creating Arch floors.

    Create a floor from the objects selected by the user, if any. Exclude
    objects that appear higher in the object hierarchy, such as sites or
    buildings. If free linking is enabled in the BIM preferences, allow higher
    hierarchy objects to be part of floors.

    Find documentation on the end user usage of Arch Floor here:
    https://wiki.freecad.org/Arch_Floor
    """

    def GetResources(self):
        """Return a dictionary with the visual aspects of the Arch Floor tool."""

        return {
            "Pixmap": "Arch_Floor",
            "MenuText": QT_TRANSLATE_NOOP("Arch_Floor", "Level"),
            "Accel": "L, V",
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_Floor",
                "Creates a Building Part object that represents a level, including selected objects",
            ),
        }

    def IsActive(self):
        """Determine whether or not the Arch Floor tool is active.

        Inactive commands are indicated by a greyed-out icon in the menus and toolbars.
        """

        return not FreeCAD.ActiveDocument is None

    def Activated(self):
        """Executed when Arch Floor is called.

        Create a floor from the objects selected by the user, if any. Exclude
        objects that appear higher in the object hierarchy, such as sites or
        buildings. If free linking is enabled in the BIM preferences, allow
        higher hierarchy objects to be part of floors.
        """

        sel = FreeCADGui.Selection.getSelection()
        link = params.get_param_arch("FreeLinking")
        floorobj = []
        warning = False
        for obj in sel:
            if not Draft.getType(obj) in ["Site", "Building"]:
                floorobj.append(obj)
            else:
                if link:
                    floorobj.append(obj)
                else:
                    warning = True
        if warning:
            message = (
                translate(
                    "Arch",
                    "You can put anything but the following objects: Site, Building, and Floor - in a Floor object.\n\
Floor object is not allowed to accept Site, Building, or Floor objects.\n\
Site, Building, and Floor objects will be removed from the selection.\n\
You can change that in the preferences.",
                )
                + "\n"
            )
            ArchCommands.printMessage(message)
        if sel and len(floorobj) == 0:
            message = (
                translate(
                    "Arch",
                    "There is no valid object in the selection.\n\
Floor creation aborted.",
                )
                + "\n"
            )
            ArchCommands.printMessage(message)
        else:
            ss = "[ "
            for o in floorobj:
                ss += "FreeCAD.ActiveDocument." + o.Name + ", "
            ss += "]"
            FreeCAD.ActiveDocument.openTransaction(translate("Arch", "Create Floor"))
            FreeCADGui.addModule("Arch")
            FreeCADGui.doCommand("obj = Arch.makeFloor(" + ss + ")")
            FreeCADGui.addModule("Draft")
            FreeCADGui.doCommand("Draft.autogroup(obj)")
            FreeCAD.ActiveDocument.commitTransaction()
            FreeCAD.ActiveDocument.recompute()


class _Floor(ArchIFC.IfcProduct):
    """Obsolete, superseded by the BuildingPart class, with IfcType set to "Building Storey".

    The Floor object.

    Turns a <App::DocumentObjectGroupPython> into a floor object, then
    takes a list of objects to own as its children.

    The floor can be based off either a group, or a python feature. Learn more
    about groups here: https://wiki.freecad.org/Std_Group

    Adds the properties of a floor, and sets its IFC type.

    Parameters
    ----------
    obj: <App::DocumentObjectGroupPython> or <App::FeaturePython>
        The object to turn into a Floor.
    """

    def __init__(self, obj):

        obj.Proxy = self
        self.Type = "Floor"
        self.Object = obj
        _Floor.setProperties(self, obj)
        self.IfcType = "Building Storey"

    def setProperties(self, obj):
        """Give the object properties unique to floors.

        Add the IFC product properties, and the floor's height and area.
        """

        ArchIFC.IfcProduct.setProperties(self, obj)
        pl = obj.PropertiesList
        if not "Height" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "Height",
                "Floor",
                QT_TRANSLATE_NOOP("App::Property", "The height of this object"),
                locked=True,
            )
        if not "Area" in pl:
            obj.addProperty(
                "App::PropertyArea",
                "Area",
                "Floor",
                QT_TRANSLATE_NOOP("App::Property", "The computed floor area of this floor"),
                locked=True,
            )
        if not hasattr(obj, "Placement"):
            # obj can be a Part Feature and already has a placement
            obj.addProperty(
                "App::PropertyPlacement",
                "Placement",
                "Base",
                QT_TRANSLATE_NOOP("App::Property", "The placement of this object"),
                locked=True,
            )

    def onDocumentRestored(self, obj):
        """Method run when the document is restored. Re-adds the properties."""

        _Floor.setProperties(self, obj)

    def dumps(self):

        return None

    def loads(self, state):

        self.Type = "Floor"

    def onChanged(self, obj, prop):
        """Method called when the object has a property changed.

        If the objects grouped under the floor object changes, recompute
        the Area property.

        Also calls ArchIFC.IfcProduct.onChanged().

        Parameters
        ----------
        prop: string
            The name of the property that has changed.
        """
        ArchIFC.IfcProduct.onChanged(self, obj, prop)

        if not hasattr(self, "Object"):
            # on restore, self.Object is not there anymore
            self.Object = obj
        if (prop == "Group") and hasattr(obj, "Area"):
            a = 0
            for o in Draft.getObjectsOfType(
                Draft.get_group_contents(obj.Group, addgroups=True), "Space"
            ):
                if hasattr(o, "Area"):
                    if hasattr(o.Area, "Value"):
                        a += o.Area.Value
                        if obj.Area.Value != a:
                            obj.Area = a

    def execute(self, obj):
        """Method run when the object is recomputed.

        Move its children if its placement has changed since the previous
        recompute. Set any child Walls and Structures to have the height of
        the floor if they have not Height value set.
        """

        # move children with this floor
        if hasattr(obj, "Placement"):
            if not hasattr(self, "OldPlacement"):
                self.OldPlacement = obj.Placement.copy()
            else:
                pl = obj.Placement.copy()
                if not DraftVecUtils.equals(pl.Base, self.OldPlacement.Base):
                    print("placement moved")
                    delta = pl.Base.sub(self.OldPlacement.Base)
                    for o in obj.Group:
                        if hasattr(o, "Placement"):
                            o.Placement.move(delta)
                    self.OldPlacement = pl
        # adjust childrens heights
        if obj.Height.Value:
            for o in obj.Group:
                if Draft.getType(o) in ["Wall", "Structure"]:
                    if not o.Height.Value:
                        o.Proxy.execute(o)

    def addObject(self, child):
        """Add the object to the floor's group.

        Parameters
        ----------
        child: <App::DocumentObject>
            The object to add to the floor's group.
        """

        if hasattr(self, "Object"):
            g = self.Object.Group
            if not child in g:
                g.append(child)
                self.Object.Group = g

    def removeObject(self, child):
        """Remove the object from the floor's group, if it's present.

        Parameters
        ----------
        child: <App::DocumentObject>
            The object to remove from the floor's group.
        """

        if hasattr(self, "Object"):
            g = self.Object.Group
            if child in g:
                g.remove(child)
                self.Object.Group = g


class _ViewProviderFloor:
    """Obsolete, superseded by the ViewProviderBuildingPart class.

    A View Provider for the Floor object.

    Parameters
    ----------
    vobj: <Gui.ViewProviderDocumentObject>
        The view provider to turn into a floor view provider.
    """

    def __init__(self, vobj):
        vobj.Proxy = self

    def getIcon(self):
        """Return the path to the appropriate icon.

        Returns
        -------
        str
            Path to the appropriate icon .svg file.
        """

        import Arch_rc

        return ":/icons/Arch_Floor_Tree.svg"

    def attach(self, vobj):
        """Add display modes' data to the coin scenegraph.

        Add each display mode as a coin node, whose parent is this view
        provider.

        Each display mode's node includes the data needed to display the object
        in that mode. This might include colors of faces, or the draw style of
        lines. This data is stored as additional coin nodes which are children
        of the display mode node.

        Do not add any new display modes.
        """

        self.Object = vobj.Object
        return

    def claimChildren(self):
        """Define which objects will appear as children in the tree view.

        Claim all the objects that appear in the floor's group.

        Returns
        -------
        list of <App::DocumentObject>s:
            The objects claimed as children.
        """

        if hasattr(self, "Object"):
            if self.Object:
                return self.Object.Group
        return []

    def dumps(self):

        return None

    def loads(self, state):

        return None

    def setupContextMenu(self, vobj, menu):
        """Add the floor specific options to the context menu.

        The context menu is the drop down menu that opens when the user right
        clicks on the floor in the tree view.

        Add a menu choice to convert the floor to an Arch Building Part with
        the ArchBuildingPart.convertFloors function.

        Parameters
        ----------
        menu: <PySide2.QtWidgets.QMenu>
            The context menu already assembled prior to this method being
            called.
        """

        from PySide import QtCore, QtGui
        import Arch_rc

        if FreeCADGui.activeWorkbench().name() != "BIMWorkbench":
            return
        action1 = QtGui.QAction(
            QtGui.QIcon(":/icons/Arch_BuildingPart.svg"), "Convert to BuildingPart", menu
        )
        QtCore.QObject.connect(action1, QtCore.SIGNAL("triggered()"), self.convertToBuildingPart)
        menu.addAction(action1)

    def convertToBuildingPart(self):
        """Converts the floor into an Arch Building Part.

        TODO: May be depreciated?
        """

        if hasattr(self, "Object"):
            import ArchBuildingPart
            from draftutils import todo

            todo.ToDo.delay(ArchBuildingPart.convertFloors, self.Object)


if FreeCAD.GuiUp:
    FreeCADGui.addCommand("Arch_Floor", _CommandFloor())



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2025 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

import unittest

from PySideWrapper import QtWidgets

from Widgets.addonmanager_widget_filter_selector import (
    WidgetFilterSelector,
    StatusFilter,
    ContentFilter,
)


class TestWidgetFilterSelector(unittest.TestCase):

    def setUp(self):
        self.window = QtWidgets.QDialog()
        self.window.setObjectName("Test Widget Filter Selector Window")
        self.wfs = WidgetFilterSelector(self.window)

    def tearDown(self):
        self.window.close()

    def test_instantiation(self):
        self.assertIsInstance(self.wfs, WidgetFilterSelector)

    def test_set_status_filter(self):
        """Explicitly setting the status filter should not emit a signal"""
        signal_caught = False

        def signal_handler():
            nonlocal signal_caught
            signal_caught = True

        self.wfs.filter_changed.connect(signal_handler)
        self.wfs.set_status_filter(StatusFilter.ANY)
        self.assertFalse(signal_caught)

    def test_set_content_filter(self):
        """Explicitly setting the content filter should not emit a signal"""
        signal_caught = False

        def signal_handler():
            nonlocal signal_caught
            signal_caught = True

        self.wfs.filter_changed.connect(signal_handler)
        self.wfs.set_contents_filter(ContentFilter.ANY)
        self.assertFalse(signal_caught)
